# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import pytest
import pathlib
from cpymad.madx import Madx
from scipy.stats import linregress
from scipy import constants as cst
import ducktrack as dtk
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed
from xtrack.beam_elements.elements import _angle_from_trig

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_constructor(test_context):
    elements = [
        xt.Drift(_context=test_context),
        xt.Marker(_context=test_context),
        xt.Multipole(_context=test_context, knl=[2, 3]),
        xt.RFMultipole(_context=test_context, knl=[2]),
        xt.Cavity(_context=test_context, voltage=3.),
        xt.SRotation(_context=test_context, angle=0),
        xt.XRotation(_context=test_context, angle=0),
        xt.YRotation(_context=test_context, angle=0),
        xt.ZetaShift(_context=test_context, dzeta=3E-4),
        xt.XYShift(_context=test_context, dx=1),
        xt.DipoleEdge(_context=test_context, h=1),
        xt.LimitRect(_context=test_context, min_x=5),
        xt.LimitRectEllipse(_context=test_context, max_x=6),
        xt.LimitEllipse(_context=test_context, a=10),
        xt.LimitRacetrack(_context=test_context, min_x=-3, max_x=4,
                           min_y=2, max_y=3, a=0.2, b=0.3),
        xt.LimitPolygon(_context=test_context, x_vertices=[1,-1,-1,1],
                        y_vertices=[1,1,-1,-1]),
        xt.Elens(_context=test_context, inner_radius=0.1),
        xt.Wire(_context=test_context, current=3.),
        xt.Exciter(_context=test_context, knl=[1], samples=[1,2,3,4],
                   sampling_frequency=1e3),
        xt.Bend(_context=test_context, length=1.),
        xt.Quadrupole(_context=test_context, length=1.),
        xt.ElectronCooler(_context=test_context,current=2.4,length=1.5,radius_e_beam=25*1e-3,
                                temp_perp=0.01,temp_long=0.001,magnetic_field=0.060) 
    ]

    # test to_dict / from_dict
    for ee in elements:
        dd = ee.to_dict()
        nee = ee.__class__.from_dict(dd, _context=test_context)
        # Check that the two objects are bitwise identical
        if not isinstance(test_context, xo.ContextCpu):
            ee.move(_context=xo.ContextCpu())
            nee.move(_context=xo.ContextCpu())
        assert (ee._xobject._buffer.buffer[ee._xobject._offset:ee._xobject._size]
                - nee._xobject._buffer.buffer[
                    nee._xobject._offset:nee._xobject._size]).sum() == 0


@pytest.mark.parametrize(
    'element_cls,args',
    [
        (xt.SRotation, {'angle': 8, 'sin_z': 0.4}),
        (xt.XRotation, {'angle': 8, 'sin_angle': 0.4}),
        (xt.YRotation, {'angle': 8, 'sin_angle': 0.4}),
        (xt.SRotation, {'angle': 8, 'cos_z': 0.4}),
        (xt.XRotation, {'angle': 8, 'cos_angle': 0.4}),
        (xt.YRotation, {'angle': 8, 'cos_angle': 0.4}),
        (xt.XRotation, {'angle': 8, 'tan_angle': 0.4}),
        (xt.YRotation, {'angle': 8, 'tan_angle': 0.4}),
    ],
)
def test_rotations_constructors_on_inconsistent_input(element_cls, args):
    with pytest.raises(ValueError):
        element_cls(**args)


@pytest.mark.parametrize(
    'cos,sin,tan,angle',
    [
        (np.cos(0.03), np.sin(0.03), np.tan(0.03), 0.03),
        (np.cos(0.01), np.sin(0.01), None, 0.01),
        (None, np.sin(0.02), np.tan(0.02), 0.02),
        (np.cos(0.03), None, np.tan(0.03), 0.03),
        (None, None, None, None),
        (0.1, None, None, None),
        (None, None, 0.1, None),
        (np.cos(0.02), np.sin(0.04), None, None),
        (None, np.sin(0.04), np.tan(0.02), None),
        (np.cos(0.02), None, np.tan(0.04), None),
    ]
)
def test__angle_from_trig(cos, sin, tan, angle):
    should_fail = angle is None

    if should_fail:
        with pytest.raises(ValueError):
            _angle_from_trig(cos, sin, tan)
    else:
        result, cos_res, sin_res, tan_res = _angle_from_trig(cos, sin, tan)

        if cos is not None:
            assert cos == cos_res
        if sin is not None:
            assert sin == sin_res
        if tan is not None:
            assert tan == tan_res

        xo.assert_allclose(angle, result, atol=1e-13)


@for_all_test_contexts
def test_backtrack(test_context):
    elements = [
        xt.Drift(_context=test_context),
        xt.Multipole(_context=test_context, knl=[2, 3]),
        xt.RFMultipole(_context=test_context, knl=[2]),
        xt.ReferenceEnergyIncrease(_context=test_context, Delta_p0c=42),
        xt.Cavity(_context=test_context, voltage=3.),
        xt.SRotation(_context=test_context, angle=4),
        xt.XRotation(_context=test_context, angle=0.3),
        xt.YRotation(_context=test_context, angle=0.7),
        xt.ZetaShift(_context=test_context, dzeta=3E-4),
        xt.XYShift(_context=test_context, dx=1),
        xt.DipoleEdge(_context=test_context, h=1),
        xt.LimitRect(_context=test_context, min_x=5),
        xt.LimitRectEllipse(_context=test_context, max_x=6),
        xt.LimitEllipse(_context=test_context, a=10),
        xt.LimitRacetrack(_context=test_context, min_x=-3, max_x=4,
                           min_y=2, max_y=3, a=0.2, b=0.3),
        xt.LimitPolygon(_context=test_context, x_vertices=[1,-1,-1,1],
                        y_vertices=[1,1,-1,-1]),
        xt.Elens(_context=test_context, inner_radius=0.1),
        xt.Exciter(_context=test_context, knl=[1], samples=[1,2,3],
                   sampling_frequency=1e3),
    ]

    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            delta=1e-2,
            zeta=1.)

    for element in elements:
        line_test = xt.Line(elements=[element])
        line_test.build_tracker(_context=test_context)

        # track forward and backward
        new_particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                               _context=test_context)
        line_test.track(new_particles)
        line_test.track(new_particles, backtrack=True)

        # assert that nothing changed
        for k in 'x,px,y,py,zeta,delta'.split(','):
            xo.assert_allclose(test_context.nparray_from_context_array(
                      getattr(new_particles, k))[0],
                      getattr(dtk_particle, k), rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_arr2ctx(test_context):
    d = xt.Drift(_context=test_context)

    a = [1., 2., 3.]
    assert type(d._arr2ctx(a)) is test_context.nplike_array_type

    a = test_context.zeros(shape=(20,), dtype=np.int64)
    assert type(d._arr2ctx(a)) is test_context.nplike_array_type
    assert (type(d._arr2ctx(a[1])) is int
            or (type(d._arr2ctx(a[1])) is test_context.nplike_array_type
                and d._arr2ctx(a[1]).shape == ()))

    a = np.array([1., 2., 3.])
    assert type(d._arr2ctx(a)) is test_context.nplike_array_type
    assert type(d._arr2ctx(a[1])) is float


@for_all_test_contexts
def test_drift(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            delta=1e-2,
            zeta=1.)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    drift = xt.Drift(_context=test_context, length=10.)
    drift.track(particles)

    dtk_drift = dtk.elements.Drift(length=10.)
    dtk_drift.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta,
                      rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_drift_exact(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            delta=1e-2,
            zeta=1.)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    drift = xt.Drift(_context=test_context, length=10.)
    line = xt.Line(elements=[drift])
    line.build_tracker(compile=False, _context=test_context)
    line.config.XTRACK_USE_EXACT_DRIFTS = True
    line.track(particles)

    dtk_drift = dtk.elements.DriftExact(length=10.)
    dtk_drift.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta,
                      rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_marker(test_context):
    dtk_particle = dtk.TestParticles(
        p0c=25.92e9,
        x=1e-3,
        px=1e-5,
        y=-2e-3,
        py=-1.5e-5,
        delta=1e-2,
        zeta=1.)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    marker = xt.Marker(_context=test_context)
    marker.track(particles)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta,
                      rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_elens(test_context):
    dtk_particle = dtk.TestParticles(
            p0c  = np.array([7000e9]),
            x    = np.array([1e-3]),
            px   = np.array([0.0]),
            y    = np.array([2.2e-3]),
            py   = np.array([0.0]),
            zeta = np.array([0.]))

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    elens = xt.Elens(_context=test_context,
                     inner_radius=1.1e-3,
                     outer_radius=2.2e-3,
                     elens_length=3.,
                     voltage=15e3,
                     current=5)

    elens.track(particles)

    dtk_elens = dtk.elements.Elens(
                   inner_radius=1.1e-3,
                   outer_radius=2.2e-3,
                   elens_length=3.,
                   voltage=15e3,
                   current=5)

    dtk_elens.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-2, atol=1e-2)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-9, atol=1e-9)


@for_all_test_contexts
def test_elens_measured_radial(test_context):
    def compute_coef(r_measured, j_measured, r_1_new, r_2_new,
                     r_1_old, r_2_old, p_order = 13):
        new_r = r_measured*(r_2_new-r_1_new)/(r_2_old-r_1_old)
        new_j = j_measured*(r_2_old-r_1_old)/(r_2_new-r_1_new)

        product = new_r*new_j

        delta_r = new_r[2]-new_r[1]

        numerator = [];
        s = len(new_r)
        for i in range(s):
            numerator.append((delta_r*max(np.cumsum(product[0:i+1]))))
        L = np.cumsum(product)
        denominator = max(L)*delta_r
        f_r = np.array(numerator/denominator)
        r_selected = new_r[new_j != 0]
        f_selected = f_r[new_j != 0]
        coef = np.polyfit(r_selected, f_selected, p_order)
        return coef


    particle_ref = xp.Particles(
                    p0c=np.array([7000e9]),
                    x=np.array([1e-3]),
                    px=np.array([0.0]),
                    y=np.array([2.2e-3]),
                    py=np.array([0.0]),
                    zeta=np.array([0.]))
    particle_test = particle_ref.copy(_context=test_context)

    # polynomial fit parameters for constant radial density
    r     = np.linspace(0.20338983,12,60)
    j     = np.append(np.append(np.linspace(0,4,20)*0,
           np.linspace(4,8,20)/np.linspace(4,8,20)), np.linspace(8,12,20)*0)
    C     = compute_coef(r, j, 1.4, 2.8, 4.0, 8.0)

    elens_radial_profile = xt.Elens(current=5, inner_radius=1.4e-3,
                outer_radius=2.8e-3, elens_length=3, voltage=10e3,
                coefficients_polynomial=C, _context=test_context)

    elens_constant = xt.Elens(current=5, inner_radius=1.4e-3,
                    outer_radius=2.8e-3, elens_length=3, voltage=10e3,
                    )

    elens_radial_profile.track(particle_test)
    elens_constant.track(particle_ref)

    particle_test.move(_context=xo.ContextCpu())

    xo.assert_allclose(particle_test.px[0], particle_ref.px[0],
                      rtol=1e-2, atol=1e-2)
    xo.assert_allclose(particle_test.py[0], particle_ref.py[0],
                      rtol=1e-2, atol=1e-2)


@for_all_test_contexts
def test_wire(test_context):
    dtk_particle = dtk.TestParticles(
            p0c =np.array([7000e9]),
            x   =np.array([1e-3]),
            px  =np.array([0.0]),
            y   =np.array([2.2e-3]),
            py  =np.array([0.0]),
            zeta=np.array([0.]))

    particles = xp.Particles(_context=test_context,
                             **dtk_particle.to_dict())


    wire = xt.Wire(_context    =  test_context,
                   L_phy  =  1.3,
                   L_int  =  1.3,
                   current=  250,
                   xma    = -8e-3,
                   yma    = -10e-3)

    wire.track(particles)

    dtk_wire = dtk.elements.Wire(
                   L_phy  =  1.3,
                   L_int  =  1.3,
                   current=  250,
                   xma    = -8e-3,
                   yma    = -10e-3)

    dtk_wire.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-9, atol=1e-9)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-9, atol=1e-9)



@for_all_test_contexts
def test_linear_transfer_first_order_taylor_map(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            zeta=2.,
            delta=2E-4)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    m0 = np.arange(6,dtype=float)
    m1 = np.ones((6,6),dtype=float)
    for i in range(6):
        for j in range(6):
            m1[i,j] = 10*i+j
    arc = xt.FirstOrderTaylorMap(_context=test_context,m0 = m0, m1 = m1)
    arc.track(particles)

    dtk_arc = dtk.elements.FirstOrderTaylorMap(m0 = m0, m1 = m1)
    dtk_arc.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_cavity(test_context):
    cav = xt.Cavity(_context=test_context, frequency=0, lag=90, voltage=30)
    part = xp.Particles(p0c=1e9, delta=[0, 1e-2], zeta=[0, 0.2], _context=test_context)
    part0 = part.copy(_context=xo.ContextCpu())

    cav.track(part)

    part = part.copy(_context=xo.ContextCpu())

    xo.assert_allclose(part.energy,
                            part0.energy+cav.voltage, atol=5e-7, rtol=0)

    Pc = np.sqrt(part.energy**2 - part.mass0**2)
    delta = Pc/part.p0c - 1
    beta = Pc/part.energy

    tau0 = part0.zeta/(part0.beta0)
    tau = part.zeta/(part.beta0)

    xo.assert_allclose(part.delta, delta, atol=1e-14, rtol=0)
    xo.assert_allclose(part.rpp, 1/(1+delta), atol=1e-14, rtol=0)
    xo.assert_allclose(part.rvv, beta/part.beta0, atol=1e-14, rtol=0)
    xo.assert_allclose(tau, tau0, atol=1e-14, rtol=0)
    xo.assert_allclose((part.ptau - part0.ptau) * part0.p0c, 30, atol=1e-9, rtol=0)

@for_all_test_contexts
def test_exciter(test_context):
    fs = 2.99792458e8 # sampling frequency in Hz
    frev = 2.99792458e8 # revolution frequency in Hz
    k0l = -0.1 # this is scaled by the waveform
    signal = [1,2,3] # this is the waveform
    duration = 4/fs
    exciter = xt.Exciter(samples=signal, sampling_frequency=fs,
                        frev=frev, start_turn=0, knl=[k0l], duration=duration)

    line = xt.Line([exciter])
    line.build_tracker(_context=test_context)

    particles = xp.Particles(p0c=6.5e12, zeta=[0,-1,-2], _context=test_context)
    num_particles = len(particles.zeta)

    line.track(particles, num_turns=1)
    expected_px = np.array([0.1, 0.2, 0.3])
    particles.move(_context=xo.context_default)

    xo.assert_allclose(particles.px, expected_px)

    particles.move(_context=test_context)
    line.track(particles, num_turns=1)
    expected_px += np.array([0.2, 0.3, 0.1])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.px, expected_px)

    particles.move(_context=test_context)
    line.track(particles, num_turns=1)
    expected_px += np.array([0.3, 0.1, 0])
    particles.move(_context=xo.context_default)
    xo.assert_allclose(particles.px, expected_px)


test_source = r"""
/*gpufun*/
void test_function(TestElementData el,
                LocalParticle* part0,
                /*gpuglmem*/ double* b){

    double const a = TestElementData_get_a(el);

    //start_per_particle_block (part0->part)

        const int64_t ipart = part->ipart;
        double const val = b[ipart];

        LocalParticle_add_to_s(part, val + a);

    //end_per_particle_block
}

/*gpufun*/
void TestElement_track_local_particle(TestElementData el,
                LocalParticle* part0){

    double const a = TestElementData_get_a(el);

    //start_per_particle_block (part0->part)

        LocalParticle_set_s(part, a);

    //end_per_particle_block
}

"""


@for_all_test_contexts
def test_per_particle_kernel(test_context):
    class TestElement(xt.BeamElement):
        _xofields = {
            'a': xo.Float64
        }

        _extra_c_sources = [test_source]

        _per_particle_kernels = {
            'test_kernel': xo.Kernel(
                c_name='test_function',
                args=[
                    xo.Arg(xo.Float64, pointer=True, name='b')
                ]),
        }

    el = TestElement(_context=test_context, a=10)

    p = xt.Particles(p0c=1e9, s=[1, 2, 3], _context=test_context)
    el.track(p)
    p.move(_context=xo.ContextCpu())
    assert np.all(p.s == [10, 10, 10])

    p = xt.Particles(p0c=1e9, s=[1, 2, 3], _context=test_context)
    b = p.s*0.5
    el.test_kernel(p, b=b)
    p.move(_context=xo.ContextCpu())
    assert np.all(p.s == np.array([11.5, 13, 14.5]))

@for_all_test_contexts
def test_simplified_accelerator_segment(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            zeta=2.,
            delta=2E-4)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    alpha_x_0 = -0.5
    beta_x_0 = 100.0
    disp_x_0 =  1.8
    disp_px_0 = 2.2
    alpha_x_1 = 2.1
    beta_x_1 = 2.0
    disp_x_1 = 3.3
    disp_px_1 = 3.7
    alpha_y_0 = -0.4
    beta_y_0 = 8.0
    disp_y_0 = -0.2
    disp_py_0 = -0.4
    alpha_y_1 = 0.7
    beta_y_1 = 0.3
    disp_y_1 = -1.9
    disp_py_1 = -2.9
    Q_x = 0.27
    Q_y = 0.34
    beta_s = 856.9
    Q_s = 0.001
    energy_ref_increment = 1.2E9
    energy_increment = 4.8E8
    x_ref_0 = -5E-3
    px_ref_0 = 6E-4
    x_ref_1 = 2E-2
    px_ref_1 = -5E-5
    y_ref_0 = -9E-2
    py_ref_0 = 1E-4
    y_ref_1 = 4E-2
    py_ref_1 = 5E-4

    arc = xt.LineSegmentMap(_context=test_context,
        alfx=(alpha_x_0, alpha_x_1), betx=(beta_x_0, beta_x_1),
        dx=(disp_x_0, disp_x_1), dpx=(disp_px_0, disp_px_1),
        alfy=(alpha_y_0, alpha_y_1), bety=(beta_y_0, beta_y_1),
        dy=(disp_y_0, disp_y_1), dpy=(disp_py_0, disp_py_1),
        qx=Q_x, qy=Q_y,
        bets=beta_s, qs=Q_s,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref=(x_ref_0, x_ref_1), px_ref=(px_ref_0, px_ref_1),
        y_ref=(y_ref_0, y_ref_1), py_ref=(py_ref_0, py_ref_1))

    arc.track(particles)

    dtk_arc = dtk.elements.LinearTransferMatrix(
        alpha_x_0=alpha_x_0, beta_x_0=beta_x_0, disp_x_0=disp_x_0, disp_px_0=disp_px_0,
        alpha_x_1=alpha_x_1, beta_x_1=beta_x_1, disp_x_1=disp_x_1, disp_px_1=disp_px_1,
        alpha_y_0=alpha_y_0, beta_y_0=beta_y_0, disp_y_0=disp_y_0, disp_py_0=disp_py_0,
        alpha_y_1=alpha_y_1, beta_y_1=beta_y_1, disp_y_1=disp_y_1, disp_py_1=disp_py_1,
        Q_x=Q_x, Q_y=Q_y,
        beta_s=beta_s, Q_s=Q_s,
        chroma_x=0.0, chroma_y=0.0,
        det_xx=0.0, det_xy=0.0, det_yy=0.0, det_yx=0.0,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref_0 = x_ref_0, px_ref_0 = px_ref_0, x_ref_1 = x_ref_1, px_ref_1 = px_ref_1,
        y_ref_0 = y_ref_0, py_ref_0 = py_ref_0, y_ref_1 = y_ref_1, py_ref_1 = py_ref_1)

    dtk_arc.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)

@for_all_test_contexts
def test_simplified_accelerator_segment_bucket(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=2e-3,
            px=4e-5,
            y=4e-3,
            py=-8e-5,
            zeta=2.,
            delta=2E-3)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)
    Q_x = 0.12
    Q_y = 0.75
    beta_s = 214.3
    Q_s = 0.0042
    bucket_length = 1E-9

    arc = xt.LineSegmentMap(_context=test_context,
        qx=Q_x, betx = 1.0, qy=Q_y, bety = 1.0,
        bets=beta_s, qs=Q_s,bucket_length=bucket_length)

    arc.track(particles)

    dtk_arc = dtk.elements.LinearTransferMatrix(
        Q_x=Q_x, Q_y=Q_y, beta_x_0 = 1.0, beta_x_1 = 1.0,
        beta_y_0 = 1.0, beta_y_1 = 1.0,
        beta_s=beta_s, Q_s=Q_s,bucket_length=bucket_length)

    dtk_arc.track(dtk_particle)

    assert np.isclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    assert np.isclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    assert np.isclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    assert np.isclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    assert np.isclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    assert np.isclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)

@for_all_test_contexts
def test_simplified_accelerator_segment_bucket_fixed_rf(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=2e-3,
            px=4e-5,
            y=4e-3,
            py=-8e-5,
            zeta=8.,
            delta=2E-3)

    particles = xp.Particles.from_dict(dtk_particle.to_dict()).copy(_context=test_context)
    Q_x = 0.12
    Q_y = 0.75
    voltage = 10E6
    f_RF = 100E6
    circumference = 2E3
    momentum_compaction = 1E-2

    arc = xt.LineSegmentMap(_context=test_context,
        qx=Q_x, betx = 1.0, qy=Q_y, bety = 1.0,
        voltage_rf = voltage,
        longitudinal_mode = 'linear_fixed_rf',
        frequency_rf = f_RF,
        lag_rf = 180.0,
        slippage_length = circumference,
        momentum_compaction_factor = momentum_compaction)

    arc.track(particles)

    particles.move(_context=xo.ContextCpu())
    eta = (momentum_compaction - 1.0 / particles.gamma0 ** 2)
    h = f_RF * circumference / (particles.beta0*cst.c)
    p0 = particles.mass0 * cst.e * particles.beta0  * particles.gamma0 / cst.c
    Q_s = np.sqrt(cst.e * voltage * eta * h / (2 * np.pi * particles.beta0 * cst.c * p0))
    beta_s = eta * circumference / (2 * np.pi * Q_s)
    Qx = 0.31
    Qy = 0.32

    dtk_arc = dtk.elements.LinearTransferMatrix(
        Q_x=Q_x, Q_y=Q_y, beta_x_0 = 1.0, beta_x_1 = 1.0,
        beta_y_0 = 1.0, beta_y_1 = 1.0,
        beta_s=beta_s, Q_s=Q_s,bucket_length=1.0/f_RF)

    dtk_arc.track(dtk_particle)

    assert np.isclose(particles.x[0], dtk_particle.x, rtol=1e-14, atol=1e-14)
    assert np.isclose(particles.px[0], dtk_particle.px, rtol=1e-14, atol=1e-14)
    assert np.isclose(particles.y[0], dtk_particle.y, rtol=1e-14, atol=1e-14)
    assert np.isclose(particles.py[0], dtk_particle.py, rtol=1e-14, atol=1e-14)
    assert np.isclose(particles.zeta[0], dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    assert np.isclose(particles.delta[0], dtk_particle.delta, rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_simplified_accelerator_segment_chroma_detuning(test_context):
    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            zeta=2.,
            delta=2E-4)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)

    alpha_x_0 = -0.5
    beta_x_0 = 100.0
    disp_x_0 = 1.8
    alpha_x_1 = 2.1
    beta_x_1 = 2.0
    disp_x_1 = 3.3
    alpha_y_0 = -0.4
    beta_y_0 = 8.0
    disp_y_0 = -0.2
    alpha_y_1 = 0.7
    beta_y_1 = 0.3
    disp_y_1 = -1.9
    Q_x = 0.27
    Q_y = 0.34
    beta_s = 856.9
    Q_s = 0.001    
    #energy_ref_increment = 1.2E9
    energy_ref_increment = 0.0 # There seems to be a bug for non-zero values
    energy_increment = 4.8E8
    x_ref_0 = -5E-3
    px_ref_0 = 6E-4
    x_ref_1 = 2E-2
    px_ref_1 = -5E-5
    y_ref_0 = -9E-2
    py_ref_0 = 1E-4
    y_ref_1 = 4E-2
    py_ref_1 = 5E-4
    chroma_x=8.0
    chroma_y=-5.0
    det_xx = 1E-3
    det_xy = -2E-4
    det_yy = -6E-4
    det_yx = 3E-3

    arc = xt.LineSegmentMap(_context=test_context,
        alfx=(alpha_x_0, alpha_x_1), betx=(beta_x_0, beta_x_1),
        dx=(disp_x_0, disp_x_1), dpx=(0.0, 0.0),
        alfy=(alpha_y_0, alpha_y_1), bety=(beta_y_0, beta_y_1),
        dy=(disp_y_0, disp_y_1), dpy=(0.0, 0.0),
        qx=Q_x, qy=Q_y,
        bets=beta_s, qs=Q_s,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref=(x_ref_0, x_ref_1), px_ref=(px_ref_0, px_ref_1),
        y_ref=(y_ref_0, y_ref_1), py_ref=(py_ref_0, py_ref_1),
        dqx=chroma_x, dqy=chroma_y,
        det_xx=det_xx, det_xy=det_xy, det_yy=det_yy, det_yx=det_yx)
    arc.track(particles)

    dtk_arc = dtk.elements.LinearTransferMatrix(alpha_x_0=alpha_x_0, beta_x_0=beta_x_0, disp_x_0=disp_x_0,
        alpha_x_1=alpha_x_1, beta_x_1=beta_x_1, disp_x_1=disp_x_1,
        alpha_y_0=alpha_y_0, beta_y_0=beta_y_0, disp_y_0=disp_y_0,
        alpha_y_1=alpha_y_1, beta_y_1=beta_y_1, disp_y_1=disp_y_1,
        Q_x=Q_x, Q_y=Q_y,
        beta_s=beta_s, Q_s=Q_s,
        chroma_x=chroma_x, chroma_y=chroma_y,
        det_xx=det_xx, det_xy=det_xy, det_yy=det_yy, det_yx=det_yx,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref_0 = x_ref_0, px_ref_0 = px_ref_0, x_ref_1 = x_ref_1, px_ref_1 = px_ref_1,
        y_ref_0 = y_ref_0, py_ref_0 = py_ref_0, y_ref_1 = y_ref_1, py_ref_1 = py_ref_1)
    dtk_arc.track(dtk_particle)

    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)


@for_all_test_contexts
def test_simplified_accelerator_segment_uncorrelated_damping(test_context):
    alpha_x_0 = -0.5
    beta_x_0 = 100.0
    disp_x_0 = 1.8
    alpha_x_1 = 2.1
    beta_x_1 = 2.0
    disp_x_1 = 3.3
    alpha_y_0 = -0.4
    beta_y_0 = 8.0
    disp_y_0 = -0.2
    alpha_y_1 = 0.7
    beta_y_1 = 0.3
    disp_y_1 = -1.9
    Q_x = 0.27
    Q_y = 0.34
    beta_s = 856.9
    Q_s = 0.001
    energy_ref_increment = 1.2E9
    energy_increment = 4.8E8
    x_ref_0 = -5E-3
    px_ref_0 = 6E-4
    x_ref_1 = 2E-2
    px_ref_1 = -5E-5
    y_ref_0 = -9E-2
    py_ref_0 = 1E-4
    y_ref_1 = 4E-2
    py_ref_1 = 5E-4
    damping_rate_x = 5E-4
    damping_rate_px = 2E-4
    damping_rate_y = 1E-3
    damping_rate_py = 7E-3
    damping_rate_zeta = 2E-3
    damping_rate_pzeta = 1E-2

    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            zeta=2.,
            delta=2E-4)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)


    arc = xt.LineSegmentMap(_context=test_context,
        alfx=(alpha_x_0, alpha_x_1), betx=(beta_x_0, beta_x_1),
        dx=(disp_x_0, disp_x_1), dpx=(0.0, 0.0),
        alfy=(alpha_y_0, alpha_y_1), bety=(beta_y_0, beta_y_1),
        dy=(disp_y_0, disp_y_1), dpy=(0.0, 0.0),
        qx=Q_x, qy=Q_y,
        bets=beta_s, qs=Q_s,
        energy_ref_increment=energy_ref_increment,
        energy_increment=energy_increment,
        x_ref=(x_ref_0, x_ref_1), px_ref=(px_ref_0, px_ref_1),
        y_ref=(y_ref_0, y_ref_1), py_ref=(py_ref_0, py_ref_1),
        damping_rate_x = damping_rate_x,damping_rate_px = damping_rate_px,
        damping_rate_y = damping_rate_y,damping_rate_py = damping_rate_py,
        damping_rate_zeta = damping_rate_zeta,damping_rate_pzeta = damping_rate_pzeta)

    arc.track(particles)

    dtk_arc = dtk.elements.LinearTransferMatrix(alpha_x_0=alpha_x_0, beta_x_0=beta_x_0, disp_x_0=disp_x_0,
        alpha_x_1=alpha_x_1, beta_x_1=beta_x_1, disp_x_1=disp_x_1,
        alpha_y_0=alpha_y_0, beta_y_0=beta_y_0, disp_y_0=disp_y_0,
        alpha_y_1=alpha_y_1, beta_y_1=beta_y_1, disp_y_1=disp_y_1,
        Q_x=Q_x, Q_y=Q_y,
        beta_s=beta_s, Q_s=Q_s,
        chroma_x=0.0, chroma_y=0.0,
        det_xx=0.0, det_xy=0.0, det_yy=0.0, det_yx=0.0,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref_0 = x_ref_0, px_ref_0 = px_ref_0, x_ref_1 = x_ref_1, px_ref_1 = px_ref_1,
        y_ref_0 = y_ref_0, py_ref_0 = py_ref_0, y_ref_1 = y_ref_1, py_ref_1 = py_ref_1,
        damping_rate_x = damping_rate_x,damping_rate_px = damping_rate_px,
        damping_rate_y = damping_rate_y,damping_rate_py = damping_rate_py,
        damping_rate_zeta = damping_rate_zeta,damping_rate_pzeta = damping_rate_pzeta)
    dtk_arc.track(dtk_particle)
    
    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)

@for_all_test_contexts
def test_simplified_accelerator_segment_correlated_damping(test_context):
    alpha_x_0 = -0.5
    beta_x_0 = 100.0
    disp_x_0 = 1.8
    alpha_x_1 = 2.1
    beta_x_1 = 2.0
    disp_x_1 = 3.3
    alpha_y_0 = -0.4
    beta_y_0 = 8.0
    disp_y_0 = -0.2
    alpha_y_1 = 0.7
    beta_y_1 = 0.3
    disp_y_1 = -1.9
    Q_x = 0.27
    Q_y = 0.34
    beta_s = 856.9
    Q_s = 0.001
    energy_ref_increment = 1.2E9
    energy_increment = 4.8E8
    x_ref_0 = -5E-3
    px_ref_0 = 6E-4
    x_ref_1 = 2E-2
    px_ref_1 = -5E-5
    y_ref_0 = -9E-2
    py_ref_0 = 1E-4
    y_ref_1 = 4E-2
    py_ref_1 = 5E-4
    damping_matrix = np.reshape(np.random.randn(36),(6,6))

    dtk_particle = dtk.TestParticles(
            p0c=25.92e9,
            x=1e-3,
            px=1e-5,
            y=-2e-3,
            py=-1.5e-5,
            zeta=2.,
            delta=2E-4)

    particles = xp.Particles.from_dict(dtk_particle.to_dict(),
                                       _context=test_context)


    arc = xt.LineSegmentMap(_context=test_context,
        alfx=(alpha_x_0, alpha_x_1), betx=(beta_x_0, beta_x_1),
        dx=(disp_x_0, disp_x_1), dpx=(0.0, 0.0),
        alfy=(alpha_y_0, alpha_y_1), bety=(beta_y_0, beta_y_1),
        dy=(disp_y_0, disp_y_1), dpy=(0.0, 0.0),
        qx=Q_x, qy=Q_y,
        bets=beta_s, qs=Q_s,
        energy_ref_increment=energy_ref_increment,
        energy_increment=energy_increment,
        x_ref=(x_ref_0, x_ref_1), px_ref=(px_ref_0, px_ref_1),
        y_ref=(y_ref_0, y_ref_1), py_ref=(py_ref_0, py_ref_1),
        damping_matrix = damping_matrix)

    arc.track(particles)

    dtk_arc = dtk.elements.LinearTransferMatrix(alpha_x_0=alpha_x_0, beta_x_0=beta_x_0, disp_x_0=disp_x_0,
        alpha_x_1=alpha_x_1, beta_x_1=beta_x_1, disp_x_1=disp_x_1,
        alpha_y_0=alpha_y_0, beta_y_0=beta_y_0, disp_y_0=disp_y_0,
        alpha_y_1=alpha_y_1, beta_y_1=beta_y_1, disp_y_1=disp_y_1,
        Q_x=Q_x, Q_y=Q_y,
        beta_s=beta_s, Q_s=Q_s,
        chroma_x=0.0, chroma_y=0.0,
        det_xx=0.0, det_xy=0.0, det_yy=0.0, det_yx=0.0,
        energy_ref_increment=energy_ref_increment,energy_increment=energy_increment,
        x_ref_0 = x_ref_0, px_ref_0 = px_ref_0, x_ref_1 = x_ref_1, px_ref_1 = px_ref_1,
        y_ref_0 = y_ref_0, py_ref_0 = py_ref_0, y_ref_1 = y_ref_1, py_ref_1 = py_ref_1,
        damping_matrix = damping_matrix)
    dtk_arc.track(dtk_particle)
    
    xo.assert_allclose(test_context.nparray_from_context_array(particles.x)[0],
                      dtk_particle.x, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.px)[0],
                      dtk_particle.px, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.y)[0],
                      dtk_particle.y, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.py)[0],
                      dtk_particle.py, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.zeta)[0],
                      dtk_particle.zeta, rtol=1e-14, atol=1e-14)
    xo.assert_allclose(test_context.nparray_from_context_array(particles.delta)[0],
                      dtk_particle.delta, rtol=1e-14, atol=1e-14)

@for_all_test_contexts
def test_simplified_accelerator_segment_uncorrelated_damping_equilibrium(test_context):
    alpha_x_0 = 0.0
    beta_x_0 = 100.0
    alpha_y_0 = 0.0
    beta_y_0 = 8.0
    Q_x = 0.18
    Q_y = 0.22
    beta_s = 856.9
    Q_s = 0.015
    damping_rate_h = 5E-4
    damping_rate_v = 1E-3
    damping_rate_s = 2E-3
    
    damping_rate_x = damping_rate_h/2
    damping_rate_px = damping_rate_h/2
    damping_rate_y = damping_rate_v/2
    damping_rate_py = damping_rate_v/2
    damping_rate_zeta = 0.0
    damping_rate_pzeta = damping_rate_s
    
    energy = 45.6
    equ_emit_x = 0.3E-9
    equ_emit_y = 1E-12
    equ_length = 3.5E-3
    equ_delta = 3.8E-4
    beta_s = equ_length/equ_delta
    equ_emit_s = equ_length*equ_delta

    
    gauss_noise_ampl_px = np.sqrt(equ_emit_x*damping_rate_h/beta_x_0)
    gauss_noise_ampl_x = beta_x_0*gauss_noise_ampl_px
    gauss_noise_ampl_py = np.sqrt(equ_emit_y*damping_rate_v/beta_y_0)
    gauss_noise_ampl_y = beta_y_0*gauss_noise_ampl_py
    gauss_noise_ampl_delta = np.sqrt(2*equ_emit_s*damping_rate_s/beta_s)
    
    npart = int(1E3)
    particles = xp.Particles(_context=test_context,
                x=np.random.randn(npart)*np.sqrt(equ_emit_x*beta_x_0),
                px=np.random.randn(npart)*np.sqrt(equ_emit_x/beta_x_0),
                y=np.random.randn(npart)*np.sqrt(equ_emit_y*beta_y_0),
                py=np.random.randn(npart)*np.sqrt(equ_emit_y/beta_y_0),
                zeta=np.random.randn(npart)*np.sqrt(equ_emit_s*beta_s),
                delta=np.random.randn(npart)*np.sqrt(equ_emit_s/beta_s),
                p0c=energy*1E9)
    particles._init_random_number_generator()

    arc = xt.LineSegmentMap(_context=test_context,
        alfx=(alpha_x_0, alpha_x_0), betx=(beta_x_0, beta_x_0),
        dx=(0.0, 0.0), dpx=(0.0, 0.0),
        alfy=(alpha_y_0, alpha_y_0), bety=(beta_y_0, beta_y_0),
        dy=(0.0, 0.0), dpy=(0.0, 0.0),
        qx=Q_x, qy=Q_y,
        bets=beta_s, qs=Q_s,
        damping_rate_x=damping_rate_x,damping_rate_px=damping_rate_px,
        damping_rate_y=damping_rate_y,damping_rate_py=damping_rate_py,
        damping_rate_zeta=0.0,damping_rate_pzeta=damping_rate_pzeta,
        gauss_noise_ampl_x = gauss_noise_ampl_x, gauss_noise_ampl_px = gauss_noise_ampl_px, 
        gauss_noise_ampl_y = gauss_noise_ampl_y, gauss_noise_ampl_py = gauss_noise_ampl_py,
        gauss_noise_ampl_zeta = 0.0, gauss_noise_ampl_pzeta = gauss_noise_ampl_delta)

    gamma_x = (1.0+alpha_x_0**2)/beta_x_0
    gamma_y = (1.0+alpha_y_0**2)/beta_y_0
    n_turns = int(1E3)
    emit_x = np.zeros(n_turns,dtype=float)
    emit_y = np.zeros_like(emit_x)
    emit_s = np.zeros_like(emit_x)
    ctx2np = test_context.nparray_from_context_array
    for turn in range(n_turns):
        arc.track(particles)
        emit_x[turn] = 0.5*np.average(ctx2np(
            gamma_x*particles.x**2+2*alpha_x_0*particles.x*particles.px
            +beta_x_0*particles.px**2))
        emit_y[turn] = 0.5*np.average(ctx2np(
            gamma_y*particles.y**2+2*alpha_y_0*particles.y*particles.py
            +beta_y_0*particles.py**2))
        emit_s[turn] = 0.5*np.average(ctx2np(particles.zeta**2/beta_s
            +beta_s*particles.delta**2))
    turns = np.arange(n_turns)
    equ_emit_x_0 = np.average(emit_x)
    equ_emit_y_0 = np.average(emit_y)
    equ_emit_s_0 = np.average(emit_s)

    xo.assert_allclose(equ_emit_x,equ_emit_x_0, rtol=1e-1, atol=1e-10)
    xo.assert_allclose(equ_emit_y,equ_emit_y_0, rtol=1e-1, atol=1e-10)
    xo.assert_allclose(equ_emit_s,equ_emit_s_0, rtol=1e-1, atol=1e-10)


@for_all_test_contexts
@fix_random_seed(3638475)
def test_simplified_accelerator_segment_correlated_noise(test_context):
    npart = int(1E6)
    scale = 1E-6
    random_matrix = np.reshape(np.random.rand(36),(6,6))
    data = np.transpose(np.random.multivariate_normal(np.zeros(6),random_matrix,npart))
    covariance_matrix = np.cov(data)

    particles = xp.Particles(_context=test_context,
                x=np.zeros(npart),
                p0c=45E9)
    particles._init_random_number_generator()

    arc = xt.LineSegmentMap(_context=test_context,
        betx=1.0, bety=1.0,bets=1.0,
        qx=0.0, qy=0.0, qs=0.0,
        gauss_noise_matrix=covariance_matrix*scale
        )

    arc.track(particles)
    data = np.zeros((6,npart))
    particles.move(_context=xo.context_default)
    data[0,:] = particles.x
    data[1,:] = particles.px
    data[2,:] = particles.y
    data[3,:] = particles.py
    data[4,:] = particles.zeta
    data[5,:] = particles.pzeta
    cov = np.cov(data)/scale
    xo.assert_allclose(cov,covariance_matrix,atol=1E-4,rtol=0.1)


@for_all_test_contexts
def test_nonlinearlens(test_context):
    mad = Madx(stdout=False)

    dr_len = 1e-11
    mad.input(f"""
    ss: sequence, l={dr_len};
        lens: nllens, at=0, cnll=0.15, knll=0.3;
        ! since in MAD-X we can't track a zero-length line, we put in
        ! this tiny drift here at the end of the sequence:
        dr: drift, at={dr_len / 2}, l={dr_len};
    endsequence;
    beam;
    use, sequence=ss;
    """)

    line = xt.Line.from_madx_sequence(mad.sequence.ss)
    line.config.XTRACK_USE_EXACT_DRIFTS = True # to be consistent with madx
    line.build_tracker(_context=test_context)

    num_p_test = 10
    x_test = np.linspace(-1e-2, 2e-2, num_p_test)
    y_test = np.linspace(-3e-2, 1e-2, num_p_test)
    px_test = np.linspace(-2e-5, 4e-5, num_p_test)
    py_test = np.linspace(-4e-5, 2e-5, num_p_test)

    p0 = xp.Particles(p0c=2e9, x=x_test, px=px_test, y=y_test, py=py_test,
                    zeta=.1, ptau=1e-3)

    part = p0.copy(_context=test_context)
    line.track(part, _force_no_end_turn_actions=True)
    part.move(_context=xo.context_default)

    xt_tau = part.zeta/part.beta0
    px = []
    py = []
    for ii in range(len(p0.x)):
        mad.input(f"""
        beam, particle=proton, pc={p0.p0c[ii] / 1e9}, sequence=ss, radiate=FALSE;

        track, onepass, onetable;
        start, x={p0.x[ii]}, px={p0.px[ii]}, y={p0.y[ii]}, py={p0.py[ii]}, \
            t={p0.zeta[ii]/p0.beta0[ii]}, pt={p0.ptau[ii]};
        run,
            turns=1,
            track_harmon=1e-15;  ! since in this test we don't care about
                ! losing particles due to t difference, we set track_harmon to
                ! something very small, to make t_max large.
        endtrack;
        """)

        mad_results = mad.table.mytracksumm[-1]

        px.append(mad_results.px)
        py.append(mad_results.py)

        xo.assert_allclose(part.x[ii], mad_results.x, atol=1e-14, rtol=0), 'x'
        xo.assert_allclose(part.px[ii], mad_results.px, atol=1e-14, rtol=0), 'px'
        xo.assert_allclose(part.y[ii], mad_results.y, atol=1e-14, rtol=0), 'y'
        xo.assert_allclose(part.py[ii], mad_results.py, atol=1e-14, rtol=0), 'py'
        xo.assert_allclose(xt_tau[ii], mad_results.t, atol=1e-14, rtol=0), 't'
        xo.assert_allclose(part.ptau[ii], mad_results.pt, atol=1e-14, rtol=0), 'pt'
        xo.assert_allclose(part.s[ii], mad_results.s, atol=1e-14, rtol=0), 's'

@for_all_test_contexts
def test_multipole_tilt_90_deg(test_context):

    m = xt.Multipole(knl=[0.1, 0], hxl=0.1, length=2, _context=test_context)
    p = xt.Particles(x = 0, y=0, delta=1, p0c=1e12, _context=test_context)
    ln = xt.Line(elements=[
        xt.SRotation(angle=-90.),
        m,
        xt.SRotation(angle=90.)])
    ln.build_tracker(_context=test_context)
    ln.track(p)

    # Check dispersion
    my = xt.Multipole(knl=[0.1, 0], hxl=0.1, rot_s_rad=np.deg2rad(-90), length=2, _context=test_context)
    py = xt.Particles(x = 0, y=0, delta=1., p0c=1e12, _context=test_context)
    my.track(py)

    p.move(_context=xo.context_default)
    py.move(_context=xo.context_default)

    xo.assert_allclose(p.x, py.x, rtol=0, atol=1e-14)
    xo.assert_allclose(p.y, py.y, rtol=0, atol=1e-14)
    xo.assert_allclose(p.px, py.px, rtol=0, atol=1e-14)
    xo.assert_allclose(p.py, py.py, rtol=0, atol=1e-14)
    xo.assert_allclose(p.zeta, py.zeta, rtol=0, atol=1e-14)
    xo.assert_allclose(p.ptau, py.ptau, rtol=0, atol=1e-14)

    # Check weak focusing
    pf = xt.Particles(x=0, y=0.3, delta=0., p0c=1e12, _context=test_context)
    pfy = pf.copy(_context=test_context)

    ln.track(pf)
    my.track(pfy)

    pf.move(_context=xo.context_default)
    pfy.move(_context=xo.context_default)

    xo.assert_allclose(pf.x, pfy.x, rtol=0, atol=1e-14)
    xo.assert_allclose(pf.y, pfy.y, rtol=0, atol=1e-14)
    xo.assert_allclose(pf.px, pfy.px, rtol=0, atol=1e-14)
    xo.assert_allclose(pf.py, pfy.py, rtol=0, atol=1e-14)
    xo.assert_allclose(pf.zeta, pfy.zeta, rtol=0, atol=1e-14)
    xo.assert_allclose(pf.ptau, pfy.ptau, rtol=0, atol=1e-14)






from cpymad.madx import Madx
import xtrack as xt
import xobjects as xo

assert_allclose = xo.assert_allclose

# def assert_allclose(a, b, **kwargs):
#     print(a, b)

test_data_folder = '../../test_data'

test_data_folder_str = str(test_data_folder)

mad1=Madx(stdout=False)
mad1.call(test_data_folder_str + '/hllhc15_thick/lhc.seq')
mad1.call(test_data_folder_str + '/hllhc15_thick/hllhc_sequence.madx')
mad1.input('beam, sequence=lhcb1, particle=proton, energy=7000;')
mad1.use('lhcb1')
mad1.input('beam, sequence=lhcb2, particle=proton, energy=7000, bv=-1;')
mad1.use('lhcb2')
mad1.call(test_data_folder_str + '/hllhc15_thick/opt_round_150_1500.madx')

mad1.globals['on_x1'] = 100 # Check kicker expressions
mad1.globals['on_sep2'] = 2 # Check kicker expressions
mad1.globals['on_x5'] = 123 # Check kicker expressions

mad1.globals['kqs.a23b1'] = 1e-4 # Check skew quad expressions
mad1.globals['kqs.a12b2'] = 1e-4 # Check skew quad expressions

mad1.globals['ksf.b1'] = 1e-3  # Check sext expressions
mad1.globals['ksf.b2'] = 1e-3  # Check sext expressions

mad1.globals['kss.a45b1'] = 1e-4 # Check skew sext expressions
mad1.globals['kss.a45b2'] = 1e-4 # Check skew sext expressions

mad1.globals['kof.a34b1'] = 3 # Check oct expressions
mad1.globals['kof.a34b2'] = 3 # Check oct expressions

mad1.globals['kcosx3.l2'] = 5 # Check skew oct expressions

mad1.globals['kcdx3.r1'] = 1e-4 # Check thin decapole expressions

mad1.globals['kcdsx3.r1'] = 1e-4 # Check thin skew decapole expressions

mad1.globals['kctx3.l1'] = 1e-5 # Check thin dodecapole expressions

mad1.globals['kctsx3.r1'] = 1e-5 # Check thin skew dodecapole expressions

mad1.input('twiss, sequence=lhcb1, table=twisslhcb1;')
mad1.input('twiss, sequence=lhcb2, table=twisslhcb2;')

twm1 = xt.Table(mad1.table.twisslhcb1)
twm2 = xt.Table(mad1.table.twisslhcb2)

collider = xt.Environment.from_madx(madx=mad1)
tw = collider.twiss(strengths=True, method='4d')

# Normal strengths
assert_allclose(twm1['k0l', 'mb.a20r8.b1:1'], tw.lhcb1['k0l', 'mb.a20r8.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k0l', 'mb.a20r8.b2:1'], tw.lhcb2['k0l', 'mb.a20r8.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k1l', 'mq.22l3.b1:1'], tw.lhcb1['k1l', 'mq.22l3.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k1l', 'mq.22l3.b2:1'], tw.lhcb2['k1l', 'mq.22l3.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k2l', 'ms.16l1.b1:1'], tw.lhcb1['k2l', 'ms.16l1.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k2l', 'ms.16l1.b2:1'], tw.lhcb2['k2l', 'ms.16l1.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k3l', 'mo.25l4.b1:1'], tw.lhcb1['k3l', 'mo.25l4.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k3l', 'mo.24l4.b2:1'], tw.lhcb2['k3l', 'mo.24l4.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k4l', 'mcdxf.3r1:1'], tw.lhcb1['k4l', 'mcdxf.3r1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k4l', 'mcdxf.3r1:1'], tw.lhcb2['k4l', 'mcdxf.3r1'], rtol=0, atol=1e-14)

assert_allclose(twm1['k5l', 'mctxf.3l1:1'], tw.lhcb1['k5l', 'mctxf.3l1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k5l', 'mctxf.3l1:1'], tw.lhcb2['k5l', 'mctxf.3l1'], rtol=0, atol=1e-14)

# Skew strengths
assert_allclose(twm1['k1sl', 'mqs.27l3.b1:1'], tw.lhcb1['k1sl', 'mqs.27l3.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k1sl', 'mqs.23l2.b2:1'], tw.lhcb2['k1sl', 'mqs.23l2.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k2sl', 'mss.28l5.b1:1'], tw.lhcb1['k2sl', 'mss.28l5.b1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k2sl', 'mss.33l5.b2:1'], tw.lhcb2['k2sl', 'mss.33l5.b2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k3sl', 'mcosx.3l2:1'], tw.lhcb1['k3sl', 'mcosx.3l2'], rtol=0, atol=1e-14)
assert_allclose(twm2['k3sl', 'mcosx.3l2:1'], tw.lhcb2['k3sl', 'mcosx.3l2'], rtol=0, atol=1e-14)

assert_allclose(twm1['k4sl', 'mcdsxf.3r1:1'], tw.lhcb1['k4sl', 'mcdsxf.3r1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k4sl', 'mcdsxf.3r1:1'], tw.lhcb2['k4sl', 'mcdsxf.3r1'], rtol=0, atol=1e-14)

assert_allclose(twm1['k5sl', 'mctsxf.3r1:1'], tw.lhcb1['k5sl', 'mctsxf.3r1'], rtol=0, atol=1e-14)
assert_allclose(twm2['k5sl', 'mctsxf.3r1:1'], tw.lhcb2['k5sl', 'mctsxf.3r1'], rtol=0, atol=1e-14)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from cpymad.madx import Madx

# hllhc15 can be found at git@github.com:lhcopt/hllhc15.git

mad = Madx()

mad.call("sequence.madx")
mad.use('lhcb1')
mad.twiss()

import xtrack as xt
import xpart as xp
import xobjects as xo

line = xt.Line.from_madx_sequence(sequence=mad.sequence.lhcb1)
for nn, ee in zip(line.element_names, line.elements):
    if nn.startswith('acs') and hasattr(ee, 'voltage'):
        assert ee.__class__.__name__ == 'Cavity'
        ee.voltage = 1e6
        ee.frequency = 400e6
particle = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1,
                        gamma0=mad.sequence.lhcb1.beam.gamma)

import json
with open('line_and_particle.json', 'w') as fid:
    json.dump({'line': line.to_dict(), 'particle': particle.to_dict()}, fid,
              cls=xo.JEncoder, indent=4)


line_with_knobs = xt.Line.from_madx_sequence(mad.sequence['lhcb1'],
                                  deferred_expressions=True
                                  )
with open('line_w_knobs_and_particle.json', 'w') as fid:
    json.dump({'line': line_with_knobs.to_dict(),
               'particle': particle.to_dict()},
            fid,
            cls=xo.JEncoder, indent=4)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import pytest
import pathlib
from cpymad.madx import Madx
from scipy.stats import linregress
from scipy import constants as cst
import ducktrack as dtk
import xobjects as xo
import xpart as xp
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed
from xtrack.beam_elements.elements import _angle_from_trig

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_ecooler_emittance(test_context):
    """Test the electron cooler by comparing the cooling rate with Betacool for LEIR.
    """
    data = np.load(test_data_folder/'electron_cooler/emittance_betacool.npz')
    emittance_betacool = data['emittance']
    time_betacool = data['time']

    gamma0 = 1.004469679
    beta0 = np.sqrt(1 - 1 / gamma0**2)
    mass0 = 193729.0248722061 * 1e6  # eV/c^2
    clight = 299792458.0
    p0c = mass0 * beta0 * gamma0  # eV/c
    q0 = 54
    particle_ref = xp.Particles(p0c=p0c,q0=q0,mass0=mass0,beta0=beta0,gamma0=gamma0)

    circumference = 78.54370266  # m
    T_per_turn = circumference/(clight*beta0)

    qx = 1.82
    qy = 2.72
    beta_x = 5
    beta_y = 5
    qs=0.005247746218929317
    bets0=-2078.673348423543

    arc = xt.LineSegmentMap(
            qx=qx, qy=qy,
            length=circumference,
            betx=beta_x,
            bety=beta_y,
            )

    arc_matching = xt.LineSegmentMap(
            qx=qx, qy=qy,
            length=circumference,
            betx=beta_x,
            bety=beta_y,
            qs=qs,
            bets=bets0)

    line_matching=xt.Line([arc_matching])
    line_matching.build_tracker()

    num_particles=int(1e2)
    sigma_dp = 5e-3    
    gemitt_x = 14e-6
    gemitt_y = 14e-6

    nemitt_x = gemitt_x*beta0*gamma0
    nemitt_y = gemitt_y*beta0*gamma0

    particles = xp.generate_matched_gaussian_bunch(
            num_particles=num_particles,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=4.2,
            particle_ref=particle_ref,
            line=line_matching,        
            )

    particles.delta = np.random.normal(loc=0.0, scale=sigma_dp, size=num_particles)
    particles.zeta = np.random.uniform(-circumference/2, circumference/2, num_particles)

    max_time_s = 1
    int_time_s = 1*1e-3
    num_turns = int((max_time_s / T_per_turn).item())
    save_interval = int((int_time_s / T_per_turn).item())

    monitor = xt.ParticlesMonitor(start_at_turn=0, stop_at_turn=1,
                            n_repetitions=int(num_turns/save_interval),
                            repetition_period=save_interval,
                            num_particles=len(particles.x))

    current = 0.6  # amperes
    cooler_length = 2.5  # m cooler length
    radius_e_beam = 25 * 1e-3
    temp_perp = 100e-3 # <E> [eV] = kb*T
    temp_long =  1e-3 # <E> [eV]
    magnetic_field = 0.075  # T for LEIR

    electron_cooler = xt.ElectronCooler(current=current,
                                    length=cooler_length,
                                    radius_e_beam=radius_e_beam,
                                    temp_perp=temp_perp, temp_long=temp_long,
                                    magnetic_field=magnetic_field)

    line = xt.Line(elements=[monitor, electron_cooler, arc],element_names=['monitor','electron_cooler','arc'])                                    
    line.particle_ref = particle_ref
    line.build_tracker()

    line.track(particles, num_turns=num_turns,
            turn_by_turn_monitor=False,with_progress=True)

    x = monitor.x[:,:,0]
    px = monitor.px[:,:,0]
    time = monitor.at_turn[:, 0, 0] * T_per_turn

    action_x = (x**2/beta_x + beta_x*px**2)
    geo_emittance_x=np.mean(action_x,axis=1)/2

    # Match Betacool and Xsuite indices to compare emittance
    valid_indices = ~np.isnan(time_betacool)
    time_betacool = time_betacool[valid_indices]
    matched_indices = [np.abs(time - tb).argmin() for tb in time_betacool]
    emittance_xsuite = geo_emittance_x[matched_indices]
    emittance_betacool = emittance_betacool[:len(emittance_xsuite)]
    
    xo.assert_allclose(emittance_xsuite, emittance_betacool, rtol=0, atol=1e-5)
   
@for_all_test_contexts
def test_ecooler_force(test_context):
    """Test the electron cooler by comparing the cooling force with Betacool for LEIR.
    """
    # Load Betacool force data
    data_betacool = np.load(test_data_folder/'electron_cooler/force_betacool.npz')
    v_diff_betacool = data_betacool['v_diff']
    force_betacool = data_betacool['force']

    beta_rel = 0.09423258405
    gamma = 1.004469679
    current = 0.6  # Amperes
    cooler_length = 2.5  # m
    radius_e_beam = 25 * 1e-3  # m
    temp_perp = 100e-3  # eV
    temp_long = 1e-3  # eV
    magnetic_field = 0.075  # T for LEIR
    mass0 = 193729.0248722061 * 1e6  # eV/c^2
    clight = 299792458.0  # m/s
    p0c = mass0 * beta_rel * gamma  # eV/c
    q0 = 54  # Charge
    beta_x = 5  # m
    emittance = 14 * 1e-6  # Initial geometric emittance (m·rad)

    # Reference particle
    particle_ref = xp.Particles(p0c=p0c, mass0=mass0, q0=q0)

    # Electron cooler
    cooler = xt.ElectronCooler(
        current=current,
        length=cooler_length,
        radius_e_beam=radius_e_beam,
        temp_perp=temp_perp,
        temp_long=temp_long,
        magnetic_field=magnetic_field,
        record_flag=1
    )  
    
    num_particles = int(1e4)
    particles = xp.Particles(
        mass0=mass0,
        p0c=p0c,
        q0=q0,
        x=np.random.normal(0, 1e-20, num_particles),
        px=np.random.normal(0, 4 * np.sqrt(emittance / beta_x), num_particles),
        y=np.random.normal(0, 1e-20, num_particles),
        py=np.random.normal(0, 1e-20, num_particles),
        delta=np.zeros(num_particles),
        zeta=np.zeros(num_particles)
    )
   
    line = xt.Line(elements=[cooler])
    line.particle_ref = particle_ref
    line.build_tracker()

    # Start internal logging for the electron cooler
    record = line.start_internal_logging_for_elements_of_type(
        xt.ElectronCooler, capacity=10000)

    line.track(particles)
    force = record.Fx
    particle_id=record.particle_id[:num_particles]
    particle_id_sort=np.argsort(particle_id)

    force=force[particle_id_sort]

    px_tot = p0c * particles.px
    beta_diff = px_tot / (mass0 * gamma)
    v_diff = beta_diff * clight
    
    sorted_indices = np.argsort(v_diff)
    v_diff = v_diff[sorted_indices]
    force = force[sorted_indices]

    # Match Betacool and Xsuite indices to compare forces
    v_diff_betacool = v_diff_betacool[~np.isnan(v_diff_betacool)]
    matching_indices = [np.abs(v_diff - vb).argmin() for vb in v_diff_betacool]

    force_xsuite = np.array([force[i] for i in matching_indices])
    force_betacool = force_betacool[:len(v_diff_betacool)]

    xo.assert_allclose(force_xsuite, force_betacool, rtol=0, atol=10)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import time
import numpy as np

from cpymad.madx import Madx

import xtrack as xt

# Import a thick sequence
mad = Madx()
mad.call('../../test_data/clic_dr/sequence.madx')
mad.use('ring')

# Makethin
mad.input(f'''
select, flag=MAKETHIN, SLICE=4, thick=false;
select, flag=MAKETHIN, pattern=wig, slice=1;
select, flag=makethin, class=rfcavity;
MAKETHIN, SEQUENCE=ring, MAKEDIPEDGE=true;
use, sequence=RING;
''')
mad.use('ring')



import numpy as np
from cpymad.madx import Madx

import xtrack as xt
import xdeps as xd

import matplotlib.pyplot as plt

plt.close('all')
qx_test = np.linspace(4.45, 4.495, 10)
qy_test = 4.4 + qx_test * 0.0
qx_mad =[]
qy_mad =[]
dqx_mad2 =[]
dqy_mad2 =[]
qx_xsuite =[]
qy_xsuite =[]
dqx_xsuite =[]
dqy_xsuite =[]
qx_ptc =[]
qy_ptc =[]
dqx_ptc =[]
dqy_ptc =[]
betx_sel_xsuite = []
betx_sel_ptc = []
bety_sel_xsuite = []
bety_sel_ptc = []
for ii, (qx, qy) in enumerate(zip(qx_test, qy_test)):

    bumper_names = ['bi1.bsw1l1.1', 'bi1.bsw1l1.2', 'bi1.bsw1l1.3', 'bi1.bsw1l1.4']
    thick_bumpers = {
    'bi1.bsw1l1.1' : {'k0_name': 'k0BI1BSW1L11'},
    'bi1.bsw1l1.2' : {'k0_name': 'k0BI1BSW1L12'},
    'bi1.bsw1l1.3' : {'k0_name': 'k0BI1BSW1L13'},
    'bi1.bsw1l1.4' : {'k0_name': 'k0BI1BSW1L14'},
    }

    mad = Madx()

    # Load model
    mad.input('''
    call, file = 'psb.seq';
    call, file = 'psb_fb_lhc.str';
    beam, particle=PROTON, pc=0.5708301551893517;
    use, sequence=psb1;
    twiss;
    ''')

    # Add shifts
    mad.input('''
    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1*;
    EALIGN, DX=-0.0057;

    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2*;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3*;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4*;
    EALIGN, DX=-0.0442;
    ''')

    # Match tunes
    mad.input(f'''
    QH = {qx};
    QV = {qy};

    MATCH, Sequence=psb1;
        VARY, NAME = kbrqf, STEP = 1e-3;
        VARY, NAME = kbrqd, STEP = 1e-3;
        GLOBAL, Q1 = QH;
        GLOBAL, Q2 = QV;
        JACOBIAN,CALLS=1000,TOLERANCE=1.0E-18,STRATEGY=3;
    ENDMATCH;
    ''')

    # # Handle bumpers (angle = 0)
    # seq = mad.sequence.psb1
    # seq.expanded_elements['bi1.bsw1l1.1'].angle = 1e-20
    # seq.expanded_elements['bi1.bsw1l1.2'].angle = 1e-20
    # seq.expanded_elements['bi1.bsw1l1.3'].angle = 1e-20
    # seq.expanded_elements['bi1.bsw1l1.4'].angle = 1e-20

    # Store bumpers length
    for nn in bumper_names:
        thick_bumpers[nn]['length'] = mad.sequence.psb1.expanded_elements[nn].l

    # Set K0 and K2 for thick bumpers
    mad.input(f'''

    l_bsw1l1.1 = {thick_bumpers['bi1.bsw1l1.1']['length']};
    l_bsw1l1.2 = {thick_bumpers['bi1.bsw1l1.2']['length']};
    l_bsw1l1.3 = {thick_bumpers['bi1.bsw1l1.3']['length']};
    l_bsw1l1.4 = {thick_bumpers['bi1.bsw1l1.4']['length']};

    BSW_K0L = 6.6E-2;
    BSW_K2L = -9.7429e-02;
    '''
    )

    mad.input('''
    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4;
    EFCOMP, DKN:={+BSW_K0L, 0, +BSW_K2L};
    !EFCOMP, DKN:={0, 0, +BSW_K2L};

    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3;
    EFCOMP, DKN:={-BSW_K0L, 0, -BSW_K2L};

    ''')

    # Twiss and ptc twiss with errors
    twmad = mad.twiss()

    beta0 = mad.sequence.psb1.beam.beta
    qx_mad.append(twmad.summary.q1)
    qy_mad.append(twmad.summary.q2)

    twmad = twmad.dframe()

    mad.input('''
    ptc_create_universe;
    ptc_create_layout, model=3, method=6, nst=5, exact=true;
    ptc_setswitch, debuglevel=0, nocavity=false, fringe=true,
                exact_mis=true, time=true, totalpath=true;
    select_ptc_normal, q1=0, q2=0;
    select_ptc_normal, dq1=1, dq2=1;
    select_ptc_normal, dq1=2, dq2=2;
    PTC_ALIGN;
    ptc_normal, closed_orbit, normal, icase=5, no=3;
    ptc_twiss, closed_orbit, table=ptc_twiss, icase=4, no=3,
                summary_table=ptc_twiss_summary;
    ptc_end;
    ''')
    twptc = mad.table.ptc_twiss.dframe()
    qx_ptc.append(mad.table.ptc_twiss.mu1[-1])
    qy_ptc.append(mad.table.ptc_twiss.mu2[-1])
    dqx_ptc.append(mad.table.normal_results.value[2] * beta0)
    dqy_ptc.append(mad.table.normal_results.value[3] * beta0)
    t_ptc = xd.Table(mad.table.ptc_twiss)
    betx_sel_ptc.append(t_ptc['betx', 'br.stscrap22:1'])
    bety_sel_ptc.append(t_ptc['bety', 'br.stscrap22:1'])

    mad.use(sequence='psb1') # wipes out the errors

    # Put back misalignments
    mad.input('''
    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.1*;
    EALIGN, DX=-0.0057;

    SELECT,FLAG=ERROR,CLEAR;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.2*;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.3*;
    SELECT,FLAG=ERROR,PATTERN=BI1.BSW1L1.4*;
    EALIGN, DX=-0.0442;
    ''')

    # K0 and K2 as strengths
    mad.input('''

    k0BI1BSW1L11 := BSW_K0L/l_bsw1l1.1;
    k0BI1BSW1L12 := -BSW_K0L/l_bsw1l1.2;
    k0BI1BSW1L13 := -BSW_K0L/l_bsw1l1.3;
    k0BI1BSW1L14 := BSW_K0L/l_bsw1l1.4;

    bi1.bsw1l1.1, k2 := BSW_K2L/l_bsw1l1.1;
    bi1.bsw1l1.2, k2 := -BSW_K2L/l_bsw1l1.2;
    bi1.bsw1l1.3, k2 := -BSW_K2L/l_bsw1l1.3;
    bi1.bsw1l1.4, k2 := BSW_K2L/l_bsw1l1.4;

    ''')

    twmad2 = mad.twiss()
    dqx_mad2.append(twmad2.summary.dq1 * beta0)
    dqy_mad2.append(twmad2.summary.dq2 * beta0)


    line = xt.Line.from_madx_sequence(
        mad.sequence.psb1,
        allow_thick=True,
        apply_madx_errors=True,
        deferred_expressions=True,
    )
    line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV,
                                gamma0=mad.sequence.psb1.beam.gamma)
    line.configure_bend_model(core='full', edge='full')
    line.build_tracker()
    line.to_json('psb_with_chicane.json')

    tw = line.twiss(method='4d')
    qx_xsuite.append(tw.qx)
    qy_xsuite.append(tw.qy)
    dqx_xsuite.append(tw.dqx)
    dqy_xsuite.append(tw.dqy)
    betx_sel_xsuite.append(tw['betx', 'br.stscrap22'])
    bety_sel_xsuite.append(tw['bety', 'br.stscrap22'])


    beta0 = line.particle_ref.beta0[0]

    plt.figure(1+ii)
    sp1 = plt.subplot(3,1,1)
    plt.plot(tw.s, tw.betx, label='xtrack')
    plt.plot(twmad.s, twmad.betx, label='madx')
    plt.plot(twptc.s, twptc.betx, label='ptc')
    plt.plot(tw.s, tw.bety, label='xtrack')
    plt.plot(twmad.s, twmad.bety, label='madx')
    plt.plot(twptc.s, twptc.bety, label='ptc')
    plt.legend()

    plt.subplot(3,1,2, sharex=sp1)
    plt.plot(tw.s, tw.dx, label='xtrack')
    plt.plot(twmad.s, twmad.dx * beta0, label='madx')
    plt.plot(tw.s, tw.dy, label='xtrack')
    plt.plot(twmad.s, twmad.dy * beta0, label='madx')

    plt.subplot(3,1,3, sharex=sp1)
    plt.plot(tw.s, tw.x, label='xtrack')
    plt.plot(twmad.s, twmad.x, label='madx')
    plt.plot(tw.s, tw.y, label='xtrack')
    plt.plot(twmad.s, twmad.y, label='madx')

    plt.legend()

plt.figure(100)
sp1 = plt.subplot(2,1,1)
# plt.plot(qx_test, qx_mad, label='madx')
plt.plot(qx_test, qx_xsuite, label='xtrack')
plt.plot(qx_test, qx_ptc, label='ptc')
plt.legend()

plt.subplot(2,1,2, sharex=sp1, sharey=sp1)
# plt.plot(qx_test, qy_mad, label='madx')
plt.plot(qx_test, qy_xsuite, label='xtrack')
plt.plot(qx_test, qy_ptc, label='ptc')
plt.legend()

plt.figure(101)
sp1 = plt.subplot(2,1,1)
plt.plot(qx_test, dqx_xsuite, label='xtrack')
plt.plot(qx_test, dqx_ptc, label='ptc')
plt.legend()

plt.subplot(2,1,2, sharex=sp1, sharey=sp1)
plt.plot(qx_test, dqy_xsuite, label='xtrack')
plt.plot(qx_test, dqy_ptc, label='ptc')
plt.legend()

plt.figure(102)
sp1 = plt.subplot(2,1,1)
plt.plot(qx_test, betx_sel_xsuite, label='xtrack')
plt.plot(qx_test, betx_sel_ptc, label='ptc')
plt.legend()

plt.subplot(2,1,2, sharex=sp1, sharey=sp1)
plt.plot(qx_test, bety_sel_xsuite, label='xtrack')
plt.plot(qx_test, bety_sel_ptc, label='ptc')
plt.legend()




plt.show()


import xtrack as xt

# Create an environment
env = xt.Environment()

# Create a line with two quadrupoles and a marker
line = env.new_line(name='myline', components=[
    env.new('q0', xt.Quadrupole, length=2.0, at=10.),
    env.new('q1', xt.Quadrupole, length=2.0, at=20.),
    env.new('m0', xt.Marker, at=40.),
    ])

tt0 = line.get_table()
tt0.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# _end_point            40            40            40

# Create a set of new elements to be placed
env.new('s1', xt.Sextupole, length=0.1, k2=0.2)
env.new('s2', xt.Sextupole, length=0.1, k2=-0.2)
env.new('m1', xt.Marker)
env.new('m2', xt.Marker)
env.new('m3', xt.Marker)

# Insert the new elements in the line
line.append(['m1', 's1', 'm2', 's2', 'm3'])

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# m1                    40            40            40
# s1                    40         40.05          40.1
# m2                  40.1          40.1          40.1
# s2                  40.1         40.15          40.2
# m3                  40.2          40.2          40.2
# _end_point          40.2          40.2          40.2

# Elements can be appended also when they are created using the class directly.
# The element name is defined contextually:
myoct = xt.Octupole(length=0.1, k3=0.3)
line.append('o1', myoct)

tt = line.get_table()
tt.show(cols=['s_start', 's_center', 's_end'])
# is:
# name             s_start      s_center         s_end
# drift_1                0           4.5             9
# q0                     9            10            11
# drift_2               11            15            19
# q1                    19            20            21
# drift_3               21          30.5            40
# m0                    40            40            40
# m1                    40            40            40
# s1                    40         40.05          40.1
# m2                  40.1          40.1          40.1
# s2                  40.1         40.15          40.2
# m3                  40.2          40.2          40.2
# o1                  40.2         40.25          40.3
# _end_point          40.3          40.3          40.3



import pandas as pd
import numpy as np
from copy import deepcopy

import xdeps as xd
import xobjects as xo
import xtrack as xt

class MultilineLegacy:

    '''
    Class to manage multiple beam lines (they can optionally share the xdeps vars).

    Parameters
    ----------
    lines: dict
        Dictionary with the lines objects
    link_vars: bool
        If True, the variables are linked between the lines.

    '''

    def __init__(self, lines: dict, link_vars=True):
        raise ValueError('MultilineLegacy is deprecated, use Environment instead')

        self.lines = {}
        self.lines.update(lines)

        line_names = list(self.lines.keys())
        self.line_names = line_names
        line_list = [self.lines[nn] for nn in line_names]
        if link_vars:
            self._var_sharing = xt.multiline_legacy.VarSharing(
                lines=line_list, names=line_names)
            self._multiline_vars = xt.line.LineVars(self)
        else:
            self._var_sharing = None

        for nn, ll in zip(line_names, line_list):
            ll._in_multiline = self
            ll._name_in_multiline = nn

        self.metadata = {}

    def to_dict(self, include_var_management=True):

        '''
        Save the multiline to a dictionary.

        Parameters
        ----------
        include_var_management: bool
            If True, the variable management data is included in the dictionary.

        Returns
        -------
        dct: dict
            The dictionary with the multiline data.
        '''

        dct = {}
        if include_var_management:
            dct['_var_manager'] = self._var_sharing.manager.dump()
            dct['_var_management_data'] = self._var_sharing.data
        dct['lines'] = {}
        for nn, ll in self.lines.items():
            dct['lines'][nn] = ll.to_dict(include_var_management=False)

        if hasattr(self, '_bb_config') and self._bb_config is not None:
            dct['_bb_config'] = {}
            for nn, vv in self._bb_config.items():
                if nn == 'dataframes':
                    dct['_bb_config'][nn] = {}
                    for kk, vv in vv.items():
                        if vv is not None:
                            dct['_bb_config'][nn][kk] = vv.to_dict()
                        else:
                            dct['_bb_config'][nn][kk] = None
                else:
                    dct['_bb_config'][nn] = vv

        dct["metadata"] = deepcopy(self.metadata)

        return dct

    @classmethod
    def from_dict(cls, dct):

        '''
        Load a multiline from a dictionary.

        Parameters
        ----------
        dct: dict
            The dictionary with the multiline data.

        Returns
        -------
        new_multiline: Multiline
            The multiline object.
        '''

        lines = {}
        for nn, ll in dct['lines'].items():
            lines[nn] = xt.Line.from_dict(ll)

        new_multiline = cls(lines=lines, link_vars=('_var_manager' in dct))

        if '_var_manager' in dct:
            for kk in dct['_var_management_data'].keys():
                new_multiline._var_sharing.data[kk].update(
                                                dct['_var_management_data'][kk])
            new_multiline._var_sharing.manager.load(dct['_var_manager'])

        if '_bb_config' in dct:
            new_multiline._bb_config = dct['_bb_config']
            for nn, vv in dct['_bb_config']['dataframes'].items():
                if vv is not None:
                    df = pd.DataFrame(vv)
                else:
                    df = None
                new_multiline._bb_config[
                    'dataframes'][nn] = df

        if "metadata" in dct:
            new_multiline.metadata = dct["metadata"]

        return new_multiline

    def to_json(self, file, indent=1, **kwargs):
        '''Save the multiline to a json file.

        Parameters
        ----------
        file: str or file-like object
            The file to save to. If a string is provided, a file is opened and
            closed. If filename ends with '.gz' file is compressed.
            If a file-like object is provided, it is used directly.
        **kwargs: dict
            Additional keyword arguments are passed to the `Line.to_dict` method.
        '''
        xt.json.dump(self.to_dict(**kwargs), file, indent=indent)

    @classmethod
    def from_json(cls, file, **kwargs):
        '''Load a multiline from a json file.

        Parameters
        ----------
        file: str or file-like object
            The file to load from. If a string is provided, a file is opened and
            closed. If the string endswith '.gz' the file is decompressed.
            If a file-like object is provided, it is used directly.
        **kwargs: dict

        Returns
        -------
        new_multiline: Multiline
            The multiline object.
        '''
        return cls.from_dict(xt.json.load(file), **kwargs)

    @classmethod
    def from_madx(cls, filename=None, madx=None, stdout=None, return_lines=False, **kwargs):
        '''
        Load a multiline from a MAD-X file.

        Parameters
        ----------
        file: str
            The MAD-X file to load from.
        **kwargs: dict
            Additional keyword arguments are passed to the `Line.from_madx_sequence`
            method.

        Returns
        -------
        new_multiline: Multiline
            The multiline object.
        '''
        return _multiline_from_madx(cls, filename=filename, madx=madx, stdout=stdout,
                             return_lines=return_lines, **kwargs)

    def copy(self):
        '''
        Returns a deep copy of the multiline.
        '''
        return self.__class__.from_dict(self.to_dict())

    def __getstate__(self):
        out = self.__dict__.copy()
        return out

    def __setstate__(self, state):
        self.__dict__.update(state)

    def build_trackers(self, _context=None, _buffer=None, **kwargs):
        '''
        Build the trackers for the lines.

        Parameters
        ----------
        _context: xobjects.Context
            The context in which the trackers are built.
        _buffer: xobjects.Buffer
            The buffer in which the trackers are built.
        **kwargs: dict
            Additional keyword arguments are passed to the `Line.build_tracker`
            method.

        '''

        for nn, ll in self.lines.items():
            ll.build_tracker(_context=_context, _buffer=_buffer, **kwargs)


    def discard_trackers(self):
        '''
        Discard the trackers associated to the lines.
        '''

        for nn, ll in self.lines.items():
            ll.discard_tracker()

    def twiss(self, lines=None, **kwargs):

        '''
        Compute the twiss parameters for the lines.

        Parameters
        ----------
        lines: list of str
            The lines for which the twiss parameters are computed. If None,
            the twiss parameters are computed for all lines.
        **kwargs: dict
            Additional keyword arguments are passed to the `Line.twiss` method.

        Returns
        -------
        out: MultiTwiss
            A MultiTwiss object containing the twiss parameters for the lines.
        '''

        for old, new in zip(['ele_start', 'ele_stop', 'ele_init', 'twiss_init'],
                            ['start', 'end', 'init_at', 'init']):
            if old in kwargs:
                raise ValueError(f'`{old}` is deprecated. Please use `{new}`.')

        out = MultiTwiss()
        if lines is None:
            lines = self.line_names

        kwargs, kwargs_per_twiss = _dispatch_twiss_kwargs(kwargs, lines)

        for ii, nn in enumerate(lines):
            this_kwargs = kwargs.copy()
            for kk in kwargs_per_twiss.keys():
                this_kwargs[kk] = kwargs_per_twiss[kk][ii]
            out[nn] = self.lines[nn].twiss(**this_kwargs)

        out._line_names = lines

        return out

    def match(self, vary, targets, restore_if_fail=True, solver=None,
              verbose=False, check_limits=True, **kwargs):

        '''
        Change a set of knobs in the beam lines in order to match assigned targets.

        Parameters
        ----------
        vary : list of str or list of Vary objects
            List of knobs to be varied. Each knob can be a string or a Vary object
            including the knob name and the step used for computing the Jacobian
            for the optimization.
        targets : list of Target objects
            List of targets to be matched.
        restore_if_fail : bool
            If True, the beamline is restored to its initial state if the matching
            fails.
        solver : str
            Solver to be used for the matching.
        check_limits : bool
            If True (default), the limits of the knobs are checked before the
            optimization. If False, if the knobs are out of limits, the optimization
            knobs are set to the limits on the first iteration.
        verbose : bool
            If True, the matching steps are printed.
        **kwargs : dict
            Additional arguments to be passed to the twiss.

        Returns
        -------
        result_info : dict
            Dictionary containing information about the matching result.

        '''

        for old, new in zip(['ele_start', 'ele_stop', 'ele_init', 'twiss_init'],
                            ['start', 'end', 'init_at', 'init']):
            if old in kwargs:
                raise ValueError(f'`{old}` is deprecated. Please use `{new}`.')

        line_names = kwargs.get('lines', self.line_names)
        kwargs, kwargs_per_twiss = _dispatch_twiss_kwargs(kwargs, line_names)
        kwargs.update(kwargs_per_twiss)

        return xt.match.match_line(self, vary, targets,
                          restore_if_fail=restore_if_fail,
                          solver=solver, check_limits=check_limits,
                          verbose=verbose, **kwargs)

    def match_knob(self, knob_name, vary, targets,
                knob_value_start=0, knob_value_end=1,
                **kwargs):

        '''
        Match a new knob in the beam line such that the specified targets are
        matched when the knob is set to the value `knob_value_end` and the
        state of the line before tha matching is recovered when the knob is
        set to the value `knob_value_start`.

        Parameters
        ----------
        knob_name : str
            Name of the knob to be matched.
        vary : list of str or list of Vary objects
            List of existing knobs to be varied.
        targets : list of Target objects
            List of targets to be matched.
        knob_value_start : float
            Value of the knob before the matching. Defaults to 0.
        knob_value_end : float
            Value of the knob after the matching. Defaults to 1.

        '''

        opt = xt.match.match_knob_line(self, vary=vary, targets=targets,
                        knob_name=knob_name, knob_value_start=knob_value_start,
                        knob_value_end=knob_value_end, **kwargs)

        return opt

    def __getitem__(self, key: str):
        if key in self.vars:
            return self.vv[key]

        if key in self.lines:
            return self.lines[key]

        raise KeyError(f'Name {key} not found')

    def __setitem__(self, key: str, value):
        if key in self.lines:
            raise ValueError(
                'Cannot create a var `{key}` using __setitem__, as there is '
                'already a line of that name in this multiline.')

        if not np.isscalar(value) and not xd.refs.is_ref(value):
            raise ValueError('Only scalars or references are allowed')

        self.vars[key] = value

    def __dir__(self):
        return list(self.lines.keys()) + object.__dir__(self)

    def __getattr__(self, key):
        if key == 'lines':
            return object.__getattribute__(self, 'lines')
        if key in self.lines:
            return self.lines[key]
        else:
            raise AttributeError(f"Multiline object has no attribute `{key}`.")

    def set(self, key, value):
        self.__setitem__(key, value)

    def get(self, key):
        return self.__getitem__(key)

    def info(self, key, limit=12):
        self.vars[key]._info(limit=limit)

    eval = xt.Line.eval
    get_expr = xt.Line.get_expr
    new_expr = xt.Line.new_expr

    @property
    def _xdeps_eval(self):
        try:
            eva_obj = self._xdeps_eval_obj
        except AttributeError:
            eva_obj = xd.madxutils.MadxEval(variables=self._xdeps_vref,
                                            functions=self.functions,
                                            elements={})
            self._xdeps_eval_obj = eva_obj

        return eva_obj

    def ref_manager(self):
        return self._var_sharing.manager

    @property
    def vars(self):
        return self._multiline_vars

    @property
    def varval(self):
        return self.vars.val

    @property
    def vv(self): # alias for varval
        return self.vars.val

    @property
    def functions(self):
        return self._xdeps_fref

    @property
    def _xdeps_vref(self):
        if self._var_sharing is not None:
            return self._var_sharing._vref

    @property
    def _xdeps_fref(self):
        if self._var_sharing is not None:
            return self._var_sharing._fref

    @property
    def _xdeps_manager(self):
        if self._var_sharing is not None:
            return self._var_sharing.manager

    def install_beambeam_interactions(self, clockwise_line, anticlockwise_line,
                                      ip_names,
                                      num_long_range_encounters_per_side,
                                      num_slices_head_on,
                                      harmonic_number, bunch_spacing_buckets,
                                      sigmaz,
                                      delay_at_ips_slots=None):

        '''
        Install beam-beam elements in the lines. Elements are inserted in the
        lines in the appropriate positions. They are not configured and are kept
        inactive.

        Parameters
        ----------
        clockwise_line: str
            Name of the line in which the beam-beam elements for the clockwise
            beam are installed.
        anticlockwise_line: xt.Line
            Name of the line in which the beam-beam elements for the
            anticlockwise beam are installed.
        ip_names: list
            The names of the IPs in the lines around which the beam-beam
            elements need to be installed.
        num_long_range_encounters_per_side: dict
            The number of long range encounters per side for each IP.
        num_slices_head_on: int
            The number of slices to be used for  the head-on beam-beam interaction.
        harmonic_number: int
            The harmonic number of the machine.
        bunch_spacing_buckets: float
            The bunch spacing in buckets.
        sigmaz: float
            The longitudinal size of the beam.
        delay_at_ips_slots: list
            Delay between the two beams in bunch slots for each IP. It specifies
            which bunch of the anticlockwise beam interacts with bunch zero of
            the clockwise beam.

        '''

        if isinstance(num_long_range_encounters_per_side, dict):
            num_long_range_encounters_per_side = [
                num_long_range_encounters_per_side[nn] for nn in ip_names]

        # Trackers need to be invalidated to add elements
        for nn, ll in self.lines.items():
            ll.unfreeze()

        if clockwise_line is not None and anticlockwise_line is not None:
            circumference_cw = self.lines[clockwise_line].get_length()
            circumference_acw = self.lines[anticlockwise_line].get_length()
            assert np.isclose(circumference_cw, circumference_acw,
                              atol=1e-4, rtol=0)

        import xfields as xf
        bb_df_cw, bb_df_acw = xf.install_beambeam_elements_in_lines(
            line_b1=self.lines.get(clockwise_line, None),
            line_b4=self.lines.get(anticlockwise_line, None),
            ip_names=ip_names,
            num_long_range_encounters_per_side=num_long_range_encounters_per_side,
            num_slices_head_on=num_slices_head_on,
            harmonic_number=harmonic_number,
            bunch_spacing_buckets=bunch_spacing_buckets,
            sigmaz_m=sigmaz, delay_at_ips_slots=delay_at_ips_slots)

        self._bb_config = {
            'dataframes': {
                'clockwise': bb_df_cw,
                'anticlockwise': bb_df_acw
            },
            'ip_names': ip_names,
            'clockwise_line': clockwise_line,
            'anticlockwise_line': anticlockwise_line,
            'bunch_spacing_buckets': bunch_spacing_buckets,
            'harmonic_number': harmonic_number
        }

    def configure_beambeam_interactions(self, num_particles,
                                    nemitt_x, nemitt_y, crab_strong_beam=True,
                                    use_antisymmetry=False,
                                    separation_bumps=None):

        '''
        Configure the beam-beam elements in the lines.

        Parameters
        ----------
        num_particles: float
            The number of particles per bunch.
        nemitt_x: float
            The normalized emittance in the horizontal plane.
        nemitt_y: float
            The normalized emittance in the vertical plane.
        crab_strong_beam: bool
            If True, crabbing of the strong beam is taken into account.
        use_antisymmetry: bool
            If True, the antisymmetry of the optics and orbit is used to compute
            the momenta of the beam-beam interaction (in the absence of the
            counter-rotating beam)
        separation_bumps: dict
            Dictionary previding the plane of the separation bump in the IPs
            where separation is present. The keys are the IP names and the
            values are the plane ("x" or "y"). This information needs to be
            provided only when use_antisymmetry is True.

        '''

        # Check that the context in which the trackers are built is on CPU
        for nn in ["clockwise", "anticlockwise"]:
            if self._bb_config[f"{nn}_line"] is None:
                continue
            line = self.lines[self._bb_config[f"{nn}_line"]]
            if not isinstance(line.tracker._context, xo.ContextCpu):
                raise ValueError(
                    "The trackers need to be built on CPU before "
                    "configuring the beam-beam elements."
                )

        if self._bb_config['dataframes']['clockwise'] is not None:
            bb_df_cw = self._bb_config['dataframes']['clockwise'].copy()
        else:
            bb_df_cw = None

        if self._bb_config['dataframes']['anticlockwise'] is not None:
            bb_df_acw = self._bb_config['dataframes']['anticlockwise'].copy()
        else:
            bb_df_acw = None

        import xfields as xf
        xf.configure_beam_beam_elements(
            bb_df_cw=bb_df_cw,
            bb_df_acw=bb_df_acw,
            line_cw=self.lines.get(self._bb_config['clockwise_line'], None),
            line_acw=self.lines.get(self._bb_config['anticlockwise_line'], None),
            num_particles=num_particles,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y,
            crab_strong_beam=crab_strong_beam,
            ip_names=self._bb_config['ip_names'],
            use_antisymmetry=use_antisymmetry,
            separation_bumps=separation_bumps)

        self.vars['beambeam_scale'] = 1.0

        for nn in ['clockwise', 'anticlockwise']:
            if self._bb_config[f'{nn}_line'] is  None: continue

            line = self.lines[self._bb_config[f'{nn}_line']]
            df = self._bb_config['dataframes'][nn]

            for bbnn in df.index:
                self.vars[f'{bbnn}_scale_strength'] = self.vars['beambeam_scale']
                line.element_refs[bbnn].scale_strength = self.vars[f'{bbnn}_scale_strength']

    def apply_filling_pattern(self, filling_pattern_cw, filling_pattern_acw,
                             i_bunch_cw, i_bunch_acw):

        '''
        Enable only he beam-beam elements corresponding to actual encounters
        for the given filling pattern and the selected bunches.

        Parameters
        ----------

        filling_pattern_cw: list or array
            The filling pattern for the clockwise beam.
        filling_pattern_acw: list or array
            The filling pattern for the anticlockwise beam.
        i_bunch_cw: int
            The index of the bunch to be simulated for the clockwise beam.
        i_bunch_acw: int
            The index of the bunch to be simulated for the anticlockwise beam.
        '''

        import xfields as xf
        apply_filling_pattern = (
            xf.config_tools.beambeam_config_tools.config_tools.apply_filling_pattern)

        apply_filling_pattern(collider=self, filling_pattern_cw=filling_pattern_cw,
                            filling_pattern_acw=filling_pattern_acw,
                            i_bunch_cw=i_bunch_cw, i_bunch_acw=i_bunch_acw)


class MultiTwiss(dict):

    def __init__(self):
        self.__dict__ = self

def _dispatch_twiss_kwargs(kwargs, lines):
    kwargs_per_twiss = {}
    for arg_name in ['start', 'end', 'init_at', 'init',
                        '_keep_initial_particles',
                        '_initial_particles', '_ebe_monitor']:
        if arg_name not in kwargs:
            continue
        if not isinstance(kwargs[arg_name], (list, tuple)):
            kwargs_per_twiss[arg_name] = len(lines) * [kwargs[arg_name]]
            kwargs.pop(arg_name)
        else:
            assert len(kwargs[arg_name]) == len(lines), \
                f'Length of {arg_name} must be equal to the number of lines'
            kwargs_per_twiss[arg_name] = list(kwargs[arg_name])
            kwargs.pop(arg_name)
    return kwargs, kwargs_per_twiss

def _multiline_from_madx(cls, filename=None, madx=None, stdout=None, return_lines=False, **kwargs):
    '''
    Load a multiline from a MAD-X file.

    Parameters
    ----------
    file: str
        The MAD-X file to load from.
    **kwargs: dict
        Additional keyword arguments are passed to the `Line.from_madx_sequence`
        method.

    Returns
    -------
    new_multiline: Multiline
        The multiline object.
    '''
    if madx is None:
        from cpymad.madx import Madx
        madx = Madx(stdout=stdout)
    if filename is not None:
        madx.call(filename)
    lines = {}
    for nn in madx.sequence.keys():
        lines[nn] = xt.Line.from_madx_sequence(
            madx.sequence[nn],
            allow_thick=True,
            deferred_expressions=True,
            **kwargs)

        lines[nn].particle_ref = xt.Particles(
            mass0=madx.sequence[nn].beam.mass*1e9,
            q0=madx.sequence[nn].beam.charge,
            gamma0=madx.sequence[nn].beam.gamma)

        if madx.sequence[nn].beam.bv == -1:
            lines[nn].twiss_default['reverse'] = True

    if return_lines:
        return lines
    else:
        out = cls(lines=lines)
        for nn in lines.keys():
            out.lines[nn].twiss_default.update(lines[nn].twiss_default)
        return out



import numpy as np
import xtrack as xt
import xobjects as xo

# Load a line and build tracker
line = xt.Line.from_json(
    '../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, q0=1, energy0=7e12)
line.cycle('ip1', inplace=True)
line.twiss_default['method'] = '4d'

# Flat machine
for nn in line.vars.get_table().rows['on_.*|corr_.*'].name:
    line.vars[nn] = 0

line['cmrskew'] = 0
line['cmiskew'] = 0
tw0 = line.twiss()

line['cmrskew'] = 0.5e-4
line['cmiskew'] = -0.3e-4

tw = line.twiss(strengths = True)

c_min_from_k1s = (0+0j) * tw.s
for ii in xt.progress_indicator.progress(range(len(tw.s))):
    c_min_from_k1s[ii] = 1 / (2*np.pi) * np.sum(tw.k1sl * np.sqrt(tw0.betx * tw0.bety)
            * np.exp(1j * 2 * np.pi * ((tw0.mux - tw0.mux[ii]) - (tw0.muy - tw0.muy[ii]))))

xo.assert_allclose(tw.c_minus_re + 1j*tw.c_minus_im, c_min_from_k1s, rtol=5e-2, atol=0)
# Check phi1
xo.assert_allclose(tw.c_minus_re + 1j*tw.c_minus_im,
    tw.c_minus * np.exp(1j * tw.c_phi1), rtol=1e-10, atol=0)
# Check phi2
xo.assert_allclose(tw.c_minus_re + 1j*tw.c_minus_im,
    tw.c_minus * np.exp(1j * (np.pi - tw.c_phi2)), rtol=3e-2, atol=0)
# Check r1
xo.assert_allclose(tw.c_r1, np.sqrt(tw.bety1 / tw.betx1), rtol=1e-5, atol=0)
# Check r2
xo.assert_allclose(tw.c_r2, np.sqrt(tw.betx2 / tw.bety2), rtol=1e-5, atol=0)

# Check c_minus
xo.assert_allclose(np.abs(tw.c_minus_re + 1j*tw.c_minus_im), tw.c_minus, rtol=1e-10, atol=0)
xo.assert_allclose(tw.c_minus_re_0, tw.c_minus_re[0], rtol=1e-10, atol=0)
xo.assert_allclose(tw.c_minus_im_0, tw.c_minus_im[0], rtol=1e-10, atol=0)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(tw.s, np.abs(tw.c_minus_re + 1j*tw.c_minus_im - c_min_from_k1s)/np.abs(c_min_from_k1s))

plt.show()


import xtrack as xt
import numpy as np
import xdeps as xd

to_test = 'line'
to_test = 'env'

env = xt.Environment()

env.vars({
    'k.1': 1.,
    'a': 2.,
    'b': '2 * a + k.1',
})

line = env.new_line([
    env.new('bb', xt.Bend, k0='2 * b', length=3+env.vars['a'] + env.vars['b'],
        h=5., ksl=[0, '3*b']),
])

for ee in [line, env]:

    # Line/Env methods (get, set, eval, get_expr, new_expr, info)
    assert ee.get('b') == 2 * 2 + 1
    assert ee.get('bb') is env.element_dict['bb']

    assert str(ee.get_expr('b')) == "((2.0 * vars['a']) + vars['k.1'])"

    assert ee.eval('3*a - sqrt(k.1)') == 5

    ne = ee.new_expr('sqrt(3*a + 3)')
    assert xd.refs.is_ref(ne)
    assert str(ne) == "f.sqrt(((3.0 * vars['a']) + 3.0))"

    ee.info('bb') # Check that it works
    ee.info('b')
    ee.info('a')

    ee.set('c', '6*a')
    assert ee.get('c') == 6 * 2

    # Line/Env containers (env[...], env.ref[...]
    assert ee['b'] == 2 * 2 + 1
    assert type(ee['bb']).__name__ == 'View'
    assert ee['bb'].__class__.__name__ == 'Bend'

    # Vars methods (get, set, eval, get_expr, new_expr, info, get_table)
    assert ee.vars.get('b') == 2 * 2 + 1

    assert str(ee.vars.get_expr('b')) == "((2.0 * vars['a']) + vars['k.1'])"

    assert ee.vars.eval('3*a - sqrt(k.1)') == 5

    ne = ee.vars.new_expr('sqrt(3*a + 3)')
    assert xd.refs.is_ref(ne)
    assert str(ne) == "f.sqrt(((3.0 * vars['a']) + 3.0))"

    ee.vars.info('b')
    ee.vars.info('a')

    ee.vars.set('d', '7*a')
    assert ee.vars.get('d') == 7 * 2

    assert xd.refs.is_ref(ee.vars['b'])

    # View methods get_expr, get_value, get_info, get_table (for now)
    assert xd.refs.is_ref(ee['bb'].get_expr('k0'))
    assert str(ee['bb'].get_expr('k0')) == "(2.0 * vars['b'])"
    assert ee['bb'].get_expr('k0')._value == 2 * (2 * 2 + 1)
    assert ee['bb'].get_value('k0') == 2 * (2 * 2 + 1)

    tt = ee['bb'].get_table()
    assert tt['value', 'k0'] == 2 * (2 * 2 + 1)
    assert tt['expr', 'k0'] == "(2.0 * vars['b'])"



import numpy as np
import xtrack as xt
from scipy.constants import c as clight
from scipy.constants import e as qe
import pathlib

from cpymad.madx import Madx


test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()


def test_fcc_ee_solenoid_correction_new_optimizer_api():
    fname = 'fccee_t'; pc_gev = 182.5

    mad = Madx(stdout=False)
    mad.call(str(test_data_folder) + '/fcc_ee/' + fname + '.seq')
    mad.beam(particle='positron', pc=pc_gev)
    mad.use('fccee_p_ring')

    line = xt.Line.from_madx_sequence(mad.sequence.fccee_p_ring, allow_thick=True,
                                    deferred_expressions=True)
    line.particle_ref = xt.Particles(mass0=xt.ELECTRON_MASS_EV,
                                    gamma0=mad.sequence.fccee_p_ring.beam.gamma)
    line.cycle('ip.4', inplace=True)
    line.append_element(element=xt.Marker(), name='ip.4.l')

    tt = line.get_table()
    bz_data_file = test_data_folder / 'fcc_ee/Bz_closed_before_quads.dat'

    line.vars['voltca1_ref'] = line.vv['voltca1']
    if 'voltca2' in line.vars.keys():
        line.vars['voltca2_ref'] = line.vv['voltca2']
    else:
        line.vars['voltca2_ref'] = 0

    line.vars['voltca1'] = 0
    line.vars['voltca2'] = 0

    import pandas as pd
    bz_df = pd.read_csv(bz_data_file, sep=r'\s+', skiprows=1, names=['z', 'Bz'])

    l_solenoid = 4.4
    ds_sol_start = -l_solenoid / 2 * np.cos(15e-3)
    ds_sol_end = +l_solenoid / 2 * np.cos(15e-3)
    ip_sol = 'ip.1'

    theta_tilt = 15e-3 # rad
    l_beam = 4.4
    l_solenoid = l_beam * np.cos(theta_tilt)
    ds_sol_start = -l_beam / 2
    ds_sol_end = +l_beam / 2
    ip_sol = 'ip.1'

    s_sol_slices = np.linspace(-l_solenoid/2, l_solenoid/2, 1001)
    bz_sol_slices = np.interp(s_sol_slices, bz_df.z, bz_df.Bz)
    bz_sol_slices[0] = 0
    bz_sol_slices[-1] = 0

    P0_J = line.particle_ref.p0c[0] * qe / clight
    brho = P0_J / qe / line.particle_ref.q0
    ks = 0.5 * (bz_sol_slices[:-1] + bz_sol_slices[1:]) / brho
    l_sol_slices = np.diff(s_sol_slices)
    s_sol_slices_entry = s_sol_slices[:-1]

    sol_slices = []
    for ii in range(len(s_sol_slices_entry)):
        sol_slices.append(xt.Solenoid(length=l_sol_slices[ii], ks=0)) # Off for now

    s_ip = tt['s', ip_sol]

    line.discard_tracker()
    line.insert_element(name='sol_start_'+ip_sol, element=xt.Marker(),
                        at_s=s_ip + ds_sol_start)
    line.insert_element(name='sol_end_'+ip_sol, element=xt.Marker(),
                        at_s=s_ip + ds_sol_end)

    sol_start_tilt = xt.YRotation(angle=-theta_tilt * 180 / np.pi)
    sol_end_tilt = xt.YRotation(angle=+theta_tilt * 180 / np.pi)
    sol_start_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))
    sol_end_shift = xt.XYShift(dx=l_solenoid/2 * np.tan(theta_tilt))

    line.element_dict['sol_start_tilt_'+ip_sol] = sol_start_tilt
    line.element_dict['sol_end_tilt_'+ip_sol] = sol_end_tilt
    line.element_dict['sol_start_shift_'+ip_sol] = sol_start_shift
    line.element_dict['sol_end_shift_'+ip_sol] = sol_end_shift

    line.element_dict['sol_entry_'+ip_sol] = xt.Solenoid(length=0, ks=0)
    line.element_dict['sol_exit_'+ip_sol] = xt.Solenoid(length=0, ks=0)

    sol_slice_names = []
    sol_slice_names.append('sol_entry_'+ip_sol)
    for ii in range(len(s_sol_slices_entry)):
        nn = f'sol_slice_{ii}_{ip_sol}'
        line.element_dict[nn] = sol_slices[ii]
        sol_slice_names.append(nn)
    sol_slice_names.append('sol_exit_'+ip_sol)

    tt = line.get_table()
    names_upstream = list(tt.rows[:'sol_start_'+ip_sol].name)
    names_downstream = list(tt.rows['sol_end_'+ip_sol:].name[:-1]) # -1 to exclude '_end_point' added by the table

    element_names = (names_upstream
                    + ['sol_start_tilt_'+ip_sol, 'sol_start_shift_'+ip_sol]
                    + sol_slice_names
                    + ['sol_end_shift_'+ip_sol, 'sol_end_tilt_'+ip_sol]
                    + names_downstream)

    line.element_names = element_names

    # re-insert the ip
    line.element_dict.pop(ip_sol)
    tt = line.get_table()
    line.insert_element(name=ip_sol, element=xt.Marker(),
            at_s = 0.5 * (tt['s', 'sol_start_'+ip_sol] + tt['s', 'sol_end_'+ip_sol]))

    line.vars['on_corr_ip.1'] = 0

    line.build_tracker()

    # Set strength
    line.vars['on_sol_'+ip_sol] = 0
    for ii in range(len(s_sol_slices_entry)):
        nn = f'sol_slice_{ii}_{ip_sol}'
        line.element_refs[nn].ks = ks[ii] * line.vars['on_sol_'+ip_sol]

    tt = line.get_table()

    tt.rows['sol_start_ip.1':'sol_end_ip.1'].show()

    line.vars['on_corr_ip.1'] = 1
    line.vars['ks0.r1'] = 0
    line.vars['ks1.r1'] = 0
    line.vars['ks2.r1'] = 0
    line.vars['ks3.r1'] = 0
    line.vars['ks4.r1'] = 0
    line.vars['ks0.l1'] = 0
    line.vars['ks1.l1'] = 0
    line.vars['ks2.l1'] = 0
    line.vars['ks3.l1'] = 0
    line.vars['ks4.l1'] = 0

    line.element_refs['qc1r1.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks0.r1']
    line.element_refs['qc2r1.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks1.r1']
    line.element_refs['qc2r2.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks2.r1']
    line.element_refs['qc1r2.1'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks3.r1']
    line.element_refs['qc1l1.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks0.l1']
    line.element_refs['qc2l1.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks1.l1']
    line.element_refs['qc2l2.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks2.l1']
    line.element_refs['qc1l2.4'].k1s = line.vars['on_corr_ip.1'] * line.vars['ks3.l1']

    line.vars['corr_k0.r1'] = 0
    line.vars['corr_k1.r1'] = 0
    line.vars['corr_k2.r1'] = 0
    line.vars['corr_k3.r1'] = 0
    line.vars['corr_k4.r1'] = 0
    line.vars['corr_k0.l1'] = 0
    line.vars['corr_k1.l1'] = 0
    line.vars['corr_k2.l1'] = 0
    line.vars['corr_k3.l1'] = 0
    line.vars['corr_k4.l1'] = 0

    line.element_refs['qc1r1.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k0.r1']
    line.element_refs['qc2r1.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k1.r1']
    line.element_refs['qc2r2.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k2.r1']
    line.element_refs['qc1r2.1'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k3.r1']
    line.element_refs['qc1l1.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k0.l1']
    line.element_refs['qc2l1.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k1.l1']
    line.element_refs['qc2l2.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k2.l1']
    line.element_refs['qc1l2.4'].k1 += line.vars['on_corr_ip.1'] * line.vars['corr_k3.l1']


    Strategy = xt.Strategy
    Teapot = xt.Teapot
    slicing_strategies = [
        Strategy(slicing=None),  # Default catch-all as in MAD-X
        Strategy(slicing=Teapot(3), element_type=xt.Bend),
        Strategy(slicing=Teapot(3), element_type=xt.CombinedFunctionMagnet),
        # Strategy(slicing=Teapot(50), element_type=xt.Quadrupole), # Starting point
        Strategy(slicing=Teapot(5), name=r'^qf.*'),
        Strategy(slicing=Teapot(5), name=r'^qd.*'),
        Strategy(slicing=Teapot(5), name=r'^qfg.*'),
        Strategy(slicing=Teapot(5), name=r'^qdg.*'),
        Strategy(slicing=Teapot(5), name=r'^ql.*'),
        Strategy(slicing=Teapot(5), name=r'^qs.*'),
        Strategy(slicing=Teapot(10), name=r'^qb.*'),
        Strategy(slicing=Teapot(10), name=r'^qg.*'),
        Strategy(slicing=Teapot(10), name=r'^qh.*'),
        Strategy(slicing=Teapot(10), name=r'^qi.*'),
        Strategy(slicing=Teapot(10), name=r'^qr.*'),
        Strategy(slicing=Teapot(10), name=r'^qu.*'),
        Strategy(slicing=Teapot(10), name=r'^qy.*'),
        Strategy(slicing=Teapot(50), name=r'^qa.*'),
        Strategy(slicing=Teapot(50), name=r'^qc.*'),
        Strategy(slicing=Teapot(20), name=r'^sy\..*'),
        Strategy(slicing=Teapot(30), name=r'^mwi\..*'),
    ]
    line.discard_tracker()
    line.slice_thick_elements(slicing_strategies=slicing_strategies)

    # Add dipole correctors
    line.insert_element(name='mcb1.r1', element=xt.Multipole(knl=[0]),
                        at='qc1r1.1_exit')
    line.insert_element(name='mcb2.r1', element=xt.Multipole(knl=[0]),
                        at='qc1r2.1_exit')
    line.insert_element(name='mcb1.l1', element=xt.Multipole(knl=[0]),
                        at='qc1l1.4_entry')
    line.insert_element(name='mcb2.l1', element=xt.Multipole(knl=[0]),
                        at='qc1l2.4_entry')

    line.vars['acb1h.r1'] = 0
    line.vars['acb1v.r1'] = 0
    line.vars['acb2h.r1'] = 0
    line.vars['acb2v.r1'] = 0
    line.vars['acb1h.l1'] = 0
    line.vars['acb1v.l1'] = 0
    line.vars['acb2h.l1'] = 0
    line.vars['acb2v.l1'] = 0

    line.element_refs['mcb1.r1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb1h.r1']
    line.element_refs['mcb2.r1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb2h.r1']
    line.element_refs['mcb1.r1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb1v.r1']
    line.element_refs['mcb2.r1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb2v.r1']
    line.element_refs['mcb1.l1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb1h.l1']
    line.element_refs['mcb2.l1'].knl[0] = line.vars['on_corr_ip.1']*line.vars['acb2h.l1']
    line.element_refs['mcb1.l1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb1v.l1']
    line.element_refs['mcb2.l1'].ksl[0] = line.vars['on_corr_ip.1']*line.vars['acb2v.l1']

    tw_thick_no_rad = line.twiss(method='4d')

    assert line.element_names[-1] == 'ip.4.l'
    assert line.element_names[0] == 'ip.4'

    opt = line.match(
        method='4d',
        start='ip.4', end='ip.4.l',
        init=tw_thick_no_rad,
        vary=xt.VaryList(['k1qf4', 'k1qf2', 'k1qd3', 'k1qd1',], step=1e-8,
        ),
        targets=[
            xt.TargetSet(at=xt.END, mux=tw_thick_no_rad.qx, muy=tw_thick_no_rad.qy, tol=1e-5),
        ]
    )
    opt.solve()

    line.vars['on_sol_ip.1'] = 0
    tw_sol_off = line.twiss(method='4d')
    line.vars['on_sol_ip.1'] = 1
    tw_sol_on = line.twiss(method='4d')

    opt_l = line.match(
        solve=False,
        method='4d', n_steps_max=30,
        start='pqc2le.4', end='ip.1', init=tw_sol_off, init_at=xt.START,
        vary=[
            xt.VaryList(['acb1h.l1', 'acb2h.l1','acb1v.l1', 'acb2v.l1'], step=1e-8, tag='corr_l'),
            xt.VaryList(['ks1.l1', 'ks2.l1', 'ks3.l1', 'ks0.l1'], step=1e-7, tag='skew_l'),
            xt.VaryList(['corr_k1.l1', 'corr_k2.l1', 'corr_k3.l1', 'corr_k0.l1'], step=1e-6, tag='normal_l'),
        ],
        targets=[
            xt.TargetSet(['x', 'y'], value=tw_sol_off, tol=1e-7, at='ip.1', tag='orbit'),
            xt.TargetSet(['px', 'py'], value=tw_sol_off, tol=1e-10, at='ip.1', tag='orbit'),
            xt.TargetRmatrix(
                        r13=0, r14=0, r23=0, r24=0, # Y-X block
                        r31=0, r32=0, r41=0, r42=0, # X-Y block,
                        start='pqc2le.4', end='ip.1', tol=1e-6, tag='coupl'),
            xt.Target('mux', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
            xt.Target('muy', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
            xt.Target('betx', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=1, tol=1e-5),
            xt.Target('bety', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=10, tol=1e-6),
            xt.Target('alfx', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),
            xt.Target('alfy', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),

        ]
    )


    for iter in range(2):
        # Orbit alone
        opt_l.disable(target=True); opt_l.disable(vary=True)
        opt_l.enable(target='orbit'); opt_l.enable(vary='corr_l'); opt_l.solve()

        # Coupling alone
        opt_l.disable(target=True); opt_l.disable(vary=True)
        opt_l.enable(target='coupl'); opt_l.enable(vary='skew_l'); opt_l.solve()

        # phase, beta and alpha alone
        opt_l.disable(target=True); opt_l.disable(vary=True)
        opt_l.enable(vary='normal_l')
        opt_l.enable(target='mu_ip'); opt_l.solve()
        opt_l.enable(target='bet_ip'); opt_l.solve()
        opt_l.enable(target='alf_ip'); opt_l.solve()

    # All together
    opt_l.enable(target=True)
    opt_l.enable(vary=True)
    opt_l.solve()


    opt_r = line.match(
        solve=False,
        method='4d', n_steps_max=30,
        start='ip.1', end='pqc2re.1', init=tw_sol_off, init_at=xt.END,
        vary=[
            xt.VaryList(['acb1h.r1', 'acb2h.r1','acb1v.r1', 'acb2v.r1'], step=1e-8, tag='corr_r'),
            xt.VaryList(['ks1.r1', 'ks2.r1', 'ks3.r1', 'ks0.r1'], step=1e-7, tag='skew_r'),
            xt.VaryList(['corr_k1.r1', 'corr_k2.r1', 'corr_k3.r1', 'corr_k0.r1'], step=1e-6, tag='normal_r'),
        ],
        targets=[
            xt.TargetSet(['x', 'y'], value=tw_sol_off, tol=1e-7, at='ip.1', tag='orbit'),
            xt.TargetSet(['px', 'py'], value=tw_sol_off, tol=1e-10, at='ip.1', tag='orbit'),
            xt.TargetRmatrix(r13=0, r14=0, r23=0, r24=0, # Y-X block
                            r31=0, r32=0, r41=0, r42=0, # X-Y block,
                            start='ip.1', end='pqc2re.1', tol=1e-6, tag='coupl'),
            xt.Target('mux', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
            xt.Target('muy', value=tw_sol_off, at='ip.1', tag='mu_ip', weight=0.1, tol=1e-6),
            xt.Target('betx', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=1, tol=1e-5),
            xt.Target('bety', value=tw_sol_off, at='ip.1', tag='bet_ip', weight=10, tol=1e-6),
            xt.Target('alfx', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),
            xt.Target('alfy', value=tw_sol_off, at='ip.1', tag='alf_ip', weight=0.1, tol=1e-4),

        ]
    )

    for iter in range(2):
        # Orbit alone
        opt_r.disable(target=True); opt_r.disable(vary=True)
        opt_r.enable(target='orbit'); opt_r.enable(vary='corr_r'); opt_r.solve()

        # Coupling alone
        opt_r.disable(target=True); opt_r.disable(vary=True)
        opt_r.enable(target='coupl'); opt_r.enable(vary='skew_r'); opt_r.solve()

        # phase, beta and alpha alone
        opt_r.disable(target=True); opt_r.disable(vary=True)
        opt_r.enable(vary='normal_r')
        opt_r.enable(target='mu_ip'); opt_r.solve()
        opt_r.enable(target='bet_ip'); opt_r.solve()
        opt_r.enable(target='alf_ip'); opt_r.solve()

    # All together
    opt_r.enable(target=True)
    opt_r.enable(vary=True)
    opt_r.solve()

    line.to_json(fname + '_with_sol_corrected.json')

    tw_sol_on_corrected = line.twiss(method='4d')

    assert_allclose = np.testing.assert_allclose

    # Check that tilt is present
    assert_allclose(tw_sol_off['kin_xprime', 'ip.1'], np.tan(0.015), atol=1e-14, rtol=0)

    # Check that solenoid introduces coupling
    assert tw_sol_on.c_minus > 1e-4

    # Check correction
    tw_chk = tw_sol_on_corrected

    assert_allclose(tw_chk['x', 'ip.1'], 0, atol=1e-8, rtol=0)
    assert_allclose(tw_chk['y', 'ip.1'], 0, atol=1e-10, rtol=0)
    assert_allclose(tw_chk['kin_xprime', 'ip.1'], tw_sol_off['kin_xprime', 'ip.1'],  atol=1e-9, rtol=0)
    assert_allclose(tw_chk['kin_yprime', 'ip.1'], 0,  atol=1e-8, rtol=0)
    assert_allclose(tw_chk['x', 'pqc2re.1'], 0, atol=5e-8, rtol=0)
    assert_allclose(tw_chk['y', 'pqc2re.1'], 0, atol=5e-8, rtol=0)
    assert_allclose(tw_chk['kin_xprime', 'pqc2re.1'], 0, atol=1e-8, rtol=0)
    assert_allclose(tw_chk['kin_yprime', 'pqc2re.1'], 0, atol=1e-8, rtol=0)
    assert_allclose(tw_chk['x', 'pqc2le.4'], 0, atol=5e-8, rtol=0)
    assert_allclose(tw_chk['y', 'pqc2le.4'], 0, atol=5e-8, rtol=0)
    assert_allclose(tw_chk['kin_xprime', 'pqc2le.4'], 0, atol=1e-8, rtol=0)
    assert_allclose(tw_chk['kin_yprime', 'pqc2le.4'], 0, atol=1e-8, rtol=0)

    assert_allclose(tw_chk['betx', 'ip.1'], tw_sol_off['betx', 'ip.1'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['bety', 'ip.1'], tw_sol_off['bety', 'ip.1'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['alfx', 'ip.1'], tw_sol_off['alfx', 'ip.1'], atol=1e-5, rtol=0)
    assert_allclose(tw_chk['alfy', 'ip.1'], tw_sol_off['alfy', 'ip.1'], atol=1e-5, rtol=0)
    assert_allclose(tw_chk['mux', 'ip.1'], tw_sol_off['mux', 'ip.1'], atol=2e-6, rtol=0)
    assert_allclose(tw_chk['muy', 'ip.1'], tw_sol_off['muy', 'ip.1'], atol=2e-6, rtol=0)

    assert_allclose(tw_chk['betx', 'pqc2re.1'], tw_sol_off['betx', 'pqc2re.1'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['bety', 'pqc2re.1'], tw_sol_off['bety', 'pqc2re.1'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['alfx', 'pqc2re.1'], tw_sol_off['alfx', 'pqc2re.1'], atol=1e-5, rtol=5e-5)
    assert_allclose(tw_chk['alfy', 'pqc2re.1'], tw_sol_off['alfy', 'pqc2re.1'], atol=1e-5, rtol=5e-5)
    assert_allclose(tw_chk['mux', 'pqc2re.1'], tw_sol_off['mux', 'pqc2re.1'], atol=2e-6, rtol=5e-5)
    assert_allclose(tw_chk['muy', 'pqc2re.1'], tw_sol_off['muy', 'pqc2re.1'], atol=2e-6, rtol=5e-5)

    assert_allclose(tw_chk['betx', 'pqc2le.4'], tw_sol_off['betx', 'pqc2le.4'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['bety', 'pqc2le.4'], tw_sol_off['bety', 'pqc2le.4'], atol=0, rtol=5e-5)
    assert_allclose(tw_chk['alfx', 'pqc2le.4'], tw_sol_off['alfx', 'pqc2le.4'], atol=1e-5, rtol=5e-5)
    assert_allclose(tw_chk['alfy', 'pqc2le.4'], tw_sol_off['alfy', 'pqc2le.4'], atol=1e-5, rtol=5e-5)
    assert_allclose(tw_chk['mux', 'pqc2le.4'], tw_sol_off['mux', 'pqc2le.4'], atol=2e-6, rtol=5e-5)
    assert_allclose(tw_chk['muy', 'pqc2le.4'], tw_sol_off['muy', 'pqc2le.4'], atol=2e-6, rtol=5e-5)

    assert tw_chk.c_minus < 1e-6
    assert_allclose(tw_chk['betx2', 'ip.1'] / tw_chk['betx', 'ip.1'], 0, atol=1e-10)
    assert_allclose(tw_chk['bety1', 'ip.1'] / tw_chk['bety', 'ip.1'], 0, atol=1e-10)
    assert_allclose(tw_chk['betx2', 'pqc2re.1'] / tw_chk['betx', 'pqc2re.1'], 0, atol=1e-10)
    assert_allclose(tw_chk['bety1', 'pqc2re.1'] / tw_chk['bety', 'pqc2re.1'], 0, atol=1e-10)
    assert_allclose(tw_chk['betx2', 'pqc2le.4'] / tw_chk['betx', 'pqc2le.4'], 0, atol=1e-10)
    assert_allclose(tw_chk['bety1', 'pqc2le.4'] / tw_chk['bety', 'pqc2le.4'], 0, atol=1e-10)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from typing import Tuple
import numpy as np

import xobjects as xo
import xtrack as xt

from .general import _print

from xobjects.struct import Struct, MetaStruct

from .line import Line, mk_class_namespace, _has_backtrack


class SerializationHeader(xo.Struct):
    """
    In a predetermined place in the buffer we have the metadata
    offset and the element type names. These have to be separate,
    because in order to rebuild TrackerData we need to first build
    ElementRefClass.
    """
    element_ref_data_offset = xo.UInt64
    reftype_names = xo.String[:]


class TrackerData:

    def __init__(
            self,
            element_dict,
            element_names,
            element_s_locations,
            line_length,
            cache=None,
            kernel_element_classes=None,
            extra_element_classes=(),
            allow_move=False,
            _context=None,
            _buffer=None,
            _offset=None,
            _no_resolve_parents=False,
    ):
        """
        Create an immutable line suitable for serialisation.

        Parameters
        ----------
        element_dict : dict
            Dictionary of elements, keyed by name.
        element_names : list
            List of element names.
        element_s_locations : list
            List of element s locations.
        line_length : float
            Length of the line.
        kernel_element_classes : list, optional
            Explicit list of classes of elements of the line; if `None`,
            will be inferred from list.
        extra_element_classes : tuple, optional
            If `kernel_element_classes` is `None`, this list will be used to augment
            the inferred list of element classes.
        """
        if _offset is not None:
            raise ValueError('`_offset` is not supported yet')

        self._element_dict = element_dict.copy()

        # Handle replicas
        for nn in list(self._element_dict.keys()):
            ee = self._element_dict[nn]
            if isinstance(ee, xt.Replica):
                self._element_dict[nn] = ee.resolve(self._element_dict)

        self._element_names = tuple(element_names)
        self._elements = tuple([self._element_dict[ee] for ee in element_names])
        self._is_backtrackable = np.all([_has_backtrack(ee, element_dict)
                                         for ee in self._elements])
        self.extra_element_classes = extra_element_classes

        # If no buffer given, try to guess it from elements, if there is no
        # common buffer, try to guess the context from elements, if there is
        # no common context, a default will be taken.
        if _buffer is None:
            common_buffer = self.common_buffer_for_elements()
            if common_buffer is not None and _context in [common_buffer.context, None]:
                _buffer = common_buffer
            if _buffer is None and _context is None:
                _context = self.common_context_for_elements()
            _buffer = _buffer or xo.get_a_buffer(context=_context, size=64)
        elif _context is not None and _buffer.context is not _context:
            raise ValueError('The given context and buffer are not compatible.')

        check_passed = self.check_elements_in_common_buffer(_buffer, allow_move=allow_move)
        if not check_passed:
            raise RuntimeError('The elements are not in the same buffer')

        line_element_classes = set(ee._XoStruct for ee in self._elements)
        self.line_element_classes = line_element_classes
        self.element_s_locations = tuple(element_s_locations)
        self.line_length = line_length
        if cache is None:
            cache = {}
        self.cache = cache

        if not kernel_element_classes:
            kernel_element_classes = (
                line_element_classes | set(extra_element_classes))
            kernel_element_classes = sorted(kernel_element_classes, key=lambda cc: cc.__name__)
        else:
            if not line_element_classes.issubset(set(kernel_element_classes)):
                raise RuntimeError(
                    f'The following classes are not in `kernel_element_classes`: '
                    f'{line_element_classes - set(kernel_element_classes)}')

        ElementRefDataClass = xt.tracker._element_ref_data_class_from_element_classes(
                                            kernel_element_classes)
        self._element_ref_data = self.build_ref_data(_buffer, ElementRefDataClass)

        # Resolve slice parents
        for nn in element_names:
            if _no_resolve_parents:
                break
            if hasattr(self._element_dict[nn], '_parent'):
                this_parent = self._element_dict[
                    self._element_dict[nn].parent_name]
                this_parent._movable = True # Force movable
                if this_parent._buffer is not self._element_dict[nn]._buffer:
                    this_parent.move(_buffer=self._element_dict[nn]._buffer)
                self._element_dict[nn]._parent = this_parent
                this_parent._movable = True
                assert self._element_dict[nn]._parent._offset == self._element_dict[nn]._xobject._parent._offset

    def common_buffer_for_elements(self):
        """If all `self.elements` elements are in the same buffer,
        returns said buffer, otherwise returns `None`."""
        common_buffer = None
        for ee in self._elements:
            if hasattr(ee, '_buffer'):
                if common_buffer is None:
                    common_buffer = ee._buffer

                if ee._buffer is not common_buffer:
                    return None

        return common_buffer

    def common_context_for_elements(self):
        """If all `self.elements` elements are on the same context,
        returns said context, otherwise returns `None`."""
        common_context = None
        for ee in self._elements:
            if hasattr(ee, '_context'):
                if common_context is None:
                    common_context = ee._context

                if ee._context is not common_context:
                    return None

        return common_context

    def to_binary(self, buffer=None) -> Tuple[xo.context.XBuffer, int]:
        """
        Return a buffer containing a binary representation of the LineFrozen,
        together with the offset to the header in the buffer.
        These two are sufficient for recreating the line.
        """

        raise NotImplementedError('This method is not supported anymore')


    def check_elements_in_common_buffer(self, buffer, allow_move=False):
        """
        Move all the elements to the common buffer, if they are not already
        there.
        """
        for nn, ee in self._element_dict.items():
            if not hasattr(ee, '_buffer'):
                continue
            if ee._buffer is not buffer:
                if allow_move:
                    ee.move(_buffer=buffer)
                else:
                    return False

        return True

    def build_ref_data(self, buffer, element_ref_data_class):
        """
        Ensure all the elements of the line are in the buffer (which will be
        created if `buffer` is equal to `None`), and write the line metadata
        to it. If the buffer is empty, the metadata will be at the beginning.
        Returns the metadata xobject.
        """

        element_ref_data = element_ref_data_class(
            elements=len(self._elements),
            names=list(self._element_names),
            _buffer=buffer,
        )

        element_ref_data.elements = [
            self._element_dict[name]._xobject for name in self.element_names
        ]

        return element_ref_data

    @classmethod
    def from_binary(
        cls,
        buffer: xo.context.XBuffer,
        header_offset: int,
        extra_element_classes: tuple = (),
    ):
        raise NotImplementedError('This method is not supported anymore')

    @property
    def elements(self):
        return self._elements

    @property
    def element_names(self):
        return self._element_names

    @property
    def _buffer(self):
        return self._element_ref_data._buffer

    @property
    def _offset(self):
        return self._element_ref_data._offset

    @property
    def _context(self):
        return self._element_ref_data._context

    def __getstate__(self):
        out = self.__dict__.copy()
        out['_element_ref_data'] = (
            self._element_ref_data._buffer, self._element_ref_data._offset)
        out['kernel_element_classes'] = [cc._DressingClass for cc in self.kernel_element_classes]
        out['line_element_classes'] = [cc._DressingClass for cc in self.line_element_classes]
        out['extra_element_classes'] = [cc._DressingClass for cc in self.extra_element_classes]
        return out

    def __setstate__(self, state):
        buffer, offset = state.pop('_element_ref_data')
        kernel_element_classes = state.pop('kernel_element_classes')
        kernel_element_classes = [cc._XoStruct for cc in kernel_element_classes]
        self.__dict__.update(state)
        self.line_element_classes = [cc._XoStruct for cc in self.line_element_classes]
        self.extra_element_classes = [cc._XoStruct for cc in self.extra_element_classes]

        element_refs_cls = xt.tracker._element_ref_data_class_from_element_classes(
                                            kernel_element_classes)
        self._element_ref_data = element_refs_cls._from_buffer(
            buffer=buffer,
            offset=offset,
        )

    @property
    def kernel_element_classes(self):
        return self._element_ref_data.elements._itemtype._reftypes

